//---------------------------------------------------------------------------
#pragma hdrstop

#include "PDFUnit.h"
#include "StringHelp.h"
#include "GraphicsHelp.h"
#include <shellapi.h>
#include "MathsHelp.h"

//---------------------------------------------------------------------------
#pragma package(smart_init)
TPDFObject *PDFObject;
//---------------------------------------------------------------------------

/*
	Code Date: 17 November 2016
	
	LEGAL STUFF
	Use this at your own risk. I am not liable for any use/misuse/failure.
	Copyright is retained by Richard Williams for the parts he has written.
	You may use and adapt this code in your own programs as you see fit. You
	retain copyright for the modifications. It may be used for commercial or
	home purposes.
	Please share the modifications.
	Please give credit where it's due.
	Please note the original author is a mechanical engineer & not a programmer,
	so relies on the BRUTE IGNORANCE & FORCE method of programming to get to the
	end. There are almost certainly quicker ways of doing the same job.


	OVERALL DESCRIPTION.

    Main Reference document used when creating this unit:
		'PDF Reference Second edition
		Adobe Portable Document Format
		Version 1.3
		Adobe Systems Incorporated'
	May be found on the Adobe Web site.
	Seems to have been written by someone who knows the standard inside out,
	but hasn't the faintest idea what the standard will be used for. As a 
	result, it is close to useless.

	There is example code showing how to use this unit lower down in these
	notes, which should work if copied as is.

    This unit is intended to act as a simple PDF creator, creating simple pages,
    no thumbnails etc. All page positioning references are in mm, related to the
    top left of the page - as per the C++ Builder Canvas but using mm rather
    than pixels. The pdf file measures everything from the bottom left of the
    page, in points (72 per inch) - as per mathematical/engineering practice.
    The conversion from one form to the other is done automatically by the
	PDFObject based on the PageHeight defined at the time the command is made.
	All mm dimensions are double. Why? Well, I have come to grief in the past
	with running out of significant digits in an application when only using
	float numbers, so now only use doubles.

    It is intended to behave in a manner similar to that of the standard
    C++ Builder 'Printer()' object. On creation of the object,  defaults are
    specified, which may be altered prior to the BeginDocument() command.
    On opening the pdf file, the binary file is created, opened and the PDF 1.3
    header is created and 2 lines written.

	Each page should be started with a 'NewPage()' command.
	If no page is open and any visible object is positioned on the page, then a
	'NewPage()' is automatically triggered. This will only happen for the first
	page.
    The current page is flushed from the buffer by the next 'NewPage()' or
    'EndDocument()' command. All items are flushed from the buffer on to the
    page in the order they were added to the buffer.
	If items overlap, items added later will be on top of earlier items.
	The page size for the current page must be defined before the next NewPage()
	or the EndDocument() command. Ideally it should be defined immediately after
	the NewPage() for page content to be in the correct position up and down the
	page. Different page sizes within the document are permitted, as are 
	different page orientations.
	If the Page size is not defined before the next NewPage(), the page inherits
	the same size as the previous page.
	It does not matter if there are multiple page size definitions for the page
	before the next NewPage() or EndDocument(). The last one will always apply,
	but any graphics/text placed on the page are positioned relative to the top
	of the page in force at the time of creation.
	If the page size is not defined in the document at all, it inherits the A4
	Portrait which the PDFObject defaults to on instantiation.
	The Author, title etc info needs to be set before the EndDocument() command,
	but can be set before the BeginDocument() command, as can all properties.
	
	When the EndDocument() command is used, the last objects for the PDF file
	are written to disc and the file closed. If the PDFObject is destroyed 
	before the EndDocument(), the EndDocument() command is called in the 
	destructor.
	
	On destruction, if the LaunchOnExit property is true, the created PDF file
	is opened with an API ShellExecute() command. This is default behaviour.


	OVERVIEW OF PDF SPEC
	The PDF specification is structured in the form of a header, a footer and
	everything in between is in the form of an object. All characters in the PDF
	file are single byte characters, NO WIDE CHARS. This is why AnsiStrings are
	used. Anything prefaced by a '%' character is a comment in the PDF file, and
	is equivalent to a '//' in C++. Each line ends in a Carriage Return
	character & a Line Feed character.

	 HEADER
	 For the PDF 1.3 spec it is always as below. The last three bytes on the
	 second line may be different, but all characters should be above ASCII 127.
	 %PDF-1.3
	 %

	 FOOTER
	 At the end of the document there will always be 5 standard objects with
	 defined object numbers. These are:
		 1)  Object 1: Catalogue
		 2)  Object 2: Outlines         At the moment no plans to expand this
		 3)  Object 3: Pages            In the PDF spec the Pages object can allow page groups.
										At the moment the PDFObject does not group pages.
		 4)  Object 4: Producer Info
		 5)  Special Object 0: Object X ref list. 
										Gives the generation and byte offset of every object from the file start.
	 Then there is the trailer, startxref, and EOF indicator

	 CONTENT
	 The content of the PDF file is in the form of an object. Each object starts
	 with a line as below:
		[objectnumber] [object generation number] obj
	 The object ends with:
		endobj
	 In between there will be various arrays (surrounded by square brackets [ ]) or
	 data dictionaries (surrounded by << and >> ) as defined by the PDF
	 specification.
	 The PDFObject creates a new PDF file, so all objects are created with
	 object generation number 0 hard coded in.
	 The objects do not need to be in numerical order within the file.

	 eg the 60x60 pixel bitmap XObject below:
		5 0 obj
		<<
		/Type /XObject                             %this object is an XObject
		/Subtype /Image                            %the XObject is an image
		/Width 60                                  %that is 60 pixels wide
		/Height 60                                 %and 60 pixels high
		/ColorSpace /DeviceRGB                     %stream is specified as rgb triplets
		/BitsPerComponent 8                        %and there are 8 bits per colour component
		/Length 10800                              %There are 10800 bytes total (60*60*3)
		>>
		stream
		{10800 bytes of data}endstream
		endobj

	 Each Page object is similar to that below:
		7 0 obj
		<< /Type /Page                             %indicates this is a page object
		/MediaBox [0 0 595.27 841.88 ]             %specifies page bottom left and top right corners in points from bottom left
		/CropBox [34.02 56.69 566.92 799.36 ]      %specifies margin in points from bottom left
		/Rotate 0                                  %view page rotated by 0 degrees
		/PZ 1.0                                    %preferred zoom scale. Doesn't seem to be followed
		/Parent 3 0 R                              &page parent is object 3, generation 0, the 'R' indicates it's an object reference rather than an object definition
		/Annots [
		11 0 R                                     %Hyperlink annotation used on this page is object 11
		]
		/Resources <<                              %Various resource objects are called up by this page
		/ProcSet [                                 %Prepare for the following to be defined on this page
		/PDF                                       %This is a pdf page
		/ImageC                                    %that uses at least one colour image
		/Text                                      %and some text
		]
		/Font <<
		/Font1 8 0 R                               %page uses Font1, and that is defined in object 8
		/Font2 9 0 R                               %page also uses Font2, and that is defined in object 9
		>>
		/XObject <<
		/Image1 5 0 R                              %page uses bitmaps stored as XObjects. Image1 is object 5
		/Image2 6 0 R                              %and Image2 is object 6
		>>
		>>
		/Contents 12 0 R                           %the page contents are stored in object 12
		>>
		endobj

	 Each page contents object is physically located 2 objects before the page 
	 object in the pdf file, and will be similar to that below:
		12 0 obj
		<< /Length 13 0 R >>                       %the length of this content stream will be specified in object 13
		stream
		0.00 0.00 0.00 RG                          %beginning of page, define pen colour black (rgb components 0.00 to 1.00)
		1.00 1.00 1.00 rg                          %define background colour white
		1 J                                        %Line Cap Style: Round. Hard coded to be like the TCanvas object.
		0 j                                        %Line Join Style: Mitred. Hard coded to be like the TCanvas object.
		0.71 w                                     %pen width in points, 0.25mm in this case.
		1.00 0.00 0.50 RG                          %pen colour for this line, 100% Red, 0% Green, 50% Blue
		340.152 799.357 m                          %move to X340.152, Y799.357 (X120.00mm, Y282.00mm) from bottom left corner of page
		56.692 289.129 l                           %line to X56.692, Y289.129 (X20.00mm, Y102.00mm)
		S                                          %fill in line.
		0.00 1.00 1.00 rg                          %Fill colour = 0% Red, 100% Green, 100% Blue = Cyan
		285.586 657.627 m
		285.586 619.664 344.916 588.888 418.103 588.888 c    %4 point bezier curve using initial pen pos, then the three co-ord pairs specified on this line
		491.291 588.888 550.621 619.664 550.621 657.627 c    % end result is an ellipse/circle
		550.621 695.591 491.291 726.366 418.103 726.366 c
		344.916 726.366 285.586 695.591 285.586 657.627 c
		B                                          %flood fill shape enclosed and colour line
		q                                          %push current graphics state on stack: must be paired with a 'Q'
		28.346 0 0 28.346 113.384 700.146 cm       %next object will have lower left corner 28.346, 28.346 and will be 113.384 x 700.146 points wide
		/Image1 Do                                 %draw Image1 in this space
		Q                                          %pop previous saved graphics state off stack
		BT                                         %Begin text (may be more than one BT on a page, but must be matched by an ET (end text), and ideally should be divided from the next BT/ET pair by some other content type
		/Font1 10 Tf                               %Use Font1, scale 10 point
		1 0 0 1 283.460 406.686 Tm                 %Locate bottom left of text at 283.46 points across, 406.686 points up from bottom left corner, with the 1 0 0 1 being the unity rotation matrix to apply to the text. The PDFObject does not permit rotated text.
		0.00 0.00 0.00 rg                          %text colour black. Note the pen & text colours are defined by the same PDF command.
		(Hello) Tj                                 %write text 'Hello'
		/Font2 20 Tf                               %Use Font2, scale 20 point
		1 0 0 1 283.460 339.994 Tm
		0.00 0.00 0.00 rg
		(World) Tj                                 %wite text 'World' some way below 'Hello' but at the same horizontal position from the left edge
		ET                                         %end text
		endstream
		endobj
	 When content is written to a page in the PDFObject, it is actually added to
	 the TStringList PageBuffer. The page buffer is written to the PDF file just
	 before each NewPage() command is implemented.

	 The page content length object will be similar to this:
		13 0 obj
		1896                                       %content stream length in bytes
		endobj

	 A font object will be as below:
		8 0 obj
		<<
		/Type /Font                                %object is a font
		/Subtype /Type1                            %of Type1
		/Name /Font1                               %it's internal name is 'Font1'
		/BaseFont /Times-BoldItalic                %and the PDF font name is 'Times-BoldItalic' which is one of the standard assumed fonts - see later.
		/Encoding /WinAnsiEncoding                 %for special characters, the Windows/Ansi coding map is used, so the copyright sign is decimal 160
		>>
		endobj

	 There are two types of hyperlink objects, a URL and a Page Link
	 URL
		16 0 obj
		<<
		/Subtype /Link
		/Rect [ 283.46 65.04 501.63 85.04 ]        %Hotspot area, bottom left to top right in points from bottom left corner of page
		/BS <</W 0 >>                              %border width 0, hard coded
		/F 4
		/A <</Type /Action /S /URI /URI (http://www.chestnutpens.co.uk) >>   %Specifies that the action from this link is a URL link, giving the web address
		>>
		endobj

	 Page Link
		17 0 obj
		<<
		/Subtype /Link
		/Rect [ 283.46 36.69 346.87 56.69 ]
		/BS <</W 0 >>
		/F 4
		/A <</Type /Action /S /GoTo /D [ 7 0 R /XYZ 10.97 595.27 0 ] >>   %Link sends user to page defined by object 7 and specifies top left screen position to be page co-ordinate 10.97,595.27, shown at zoom 0 (current zoom)
		>>
		endobj

	IMPLEMENTATION
	Where possible when writing to a page, the names, and way the functions
	operate is intended to be the same as on a TCanvas. There is one serious
	exception though. All dimensions are in mm, which may have fractional
	components, so are defined as doubles. The 'coordinate' struct is the double
	equivalent of the TPoint structure and an array of this will be needed for
	the PolyBezier/Polygon/PolyLine functions.
    All dimensions in the document are rounded to 3 decimal places of point -
    so at 72 points per inch, this gives a real world precision of 0.0003528mm.
    This should be adequate precision even for a 2400dpi printer (comes out as
	accurate to the nearest 0.84 dot).


	COMPRESSION
	Compression of data streams is not carried out. I simply do not understand
	how to do it. The PDF file produced by this object is large, but as it's all
	text with a large number of repeating elements in it, it zips up much better
	than many files, creating small, easily distributed zip files. The PDF files
	seem to open very fast without the compression, too.

	MARGINS
	Try to avoid using the margins in the PDF object. They are used to create
	the 'CropBox' in the pdf file. This literally crops the displayed part of
	the page to the cropbox and does not display or render any part of the page
	outside that area. The margin is only visible when printing as a greyed area
	around the printed bit. You should really manage the margins inside the
	calling application, as you would with the C++ Builder Printer() object.


    BITMAPS
    When specifying a bitmap object to be saved in the pdf file, it is always
    saved as an XObject, rather than as an in-line object. The Draw command is
    an all purpose Draw/StretchDraw command, in that the bitmap may be stretched
    or distorted simply by specifying the dimensions over which it is to paint.
    If you are likely to re-use the bitmap, keep a record of the objectnumber
    returned by the Draw() function as this will be needed for the
    UseResourceImage() method to re-use it.
    The image can be saved in the pdf file before any pages are created, using
    the AddResourceImage() command. But the Draw() and UseResourceImage()
    commands can only be fully effective after the first NewPage() command.
    A Draw() or AddResourceImage() must occur before the UseResourceImage() -
	simply put, you cannot use/re-use something before defining it. If you try
	to use a bitmap before definition, the command is simply ignored.


    TEXT
    With text, the text position is, as per TCanvas practice, defined at the
    upper left corner.
	The PDF file defines the active point of the text as the lower left corner.

	Two TextHeight functions are offered. 
	The first is the 'Official' height of
	the font, which is the point size, so 14 point is 14/2.8346mm = 4.9390mm
    Thus when positioning the text, the vertical position will be defined as
    the defined position + font height down from the top left corner of the
	page, then transposed to a height up from the bottom left corner. This allows
	you to define line spacing, as it is simply the height of the font.

	The second TextHeight gives that actual height of the text presented, and is
	calculated using the same kludge as the TextWidth below.

	Text Width is a kludge. A Bitmap with a Canvas is created. Then the width in
	pixels of the text at 10x its font size is found. this is then divided by
	the number of pixels forming the text width of Arial 100point 'Hi There' and
	multiplied by the actual width of Arial 10pt 'Hi There', which is 13.517mm.

	There are 14 standard fonts that are assumed to be on each machine, and will
    be treated as Type 1 fonts which do not need to be embedded. These are:
        Courier
        Courier-Bold
        Courier-Oblique
        Courier-BoldOblique
        Helvetica
        Helvetica-Bold
        Helvetica-Oblique
        Helvetica-BoldOblique
        Times-Roman
        Times-Bold
        Times-Italic
        Times-BoldItalic
        Symbol
        ZapfDingbats
	At the moment fonts are not embedded, so you should use the standard fonts.
	On Windows machines there are standard substitutions for the above fonts, as
	below, and there is no need to be concerned with their presence:
		Courier New
		Arial
		Times New Roman
	If you do choose to use True Type Fonts, in effect, it's assumed they will
	be on the machine, leaving any substitutions down to the PDF reader app. I
	have spent many frustrating hours trying to interrogate the font system in
	the Windows API for the data to define a font glyph width etc, and got
	nowhere. As a result, TrueType fonts will always come up with a 'FontBox
	Error' when used in the PDF file, and the text size/width will be wrong.
	Please DO NOT USE TRUETYPE FONTS, JUST USE THE PRE-DEFINED FONTS.

	The PDFObject does not have a TFont object for the current font in use. The
	reason for this is that the default PDF font names probably have no windows
	equivalents. It is likely there will be 'Arial' & no 'Helvetica'. C++Builder
	10 seems to have a very strong affinity between fonts on the machine and
	fonts available in a TFont object. Giving a font name of 'Helvetica' when
	there was no such font on the machine caused the TFont object to die - I
	think it's because no Handle can be generated for a font that's not there.
	So,	instead, as the only info needed for a font is the Name and Size, a
	TPDFFont structure was created. This has a FontStyle built in, but is for
	internal use only - so don't access it directly please.

	Underlining and strikeout of the font is achieved by drawing lines through
	and/or under the text and not as an integral feature of the font. The lines
	are drawn 3/4 the way down for strikeout, and 1/10th the font height below
	the font line for underline. The line width is 1/15th the font height. This
	looks right.



    The following TCanvas methods and their equivalents are given below:-
		TCanvas             TPDFObject
		 AngleArc            AngleArc
		 Arc                 Arc
		 ArcTo               Arc, according to the C++ Builder documentation,
											this is functionally identical to
											Arc, so use Arc.
         Brush Copy          None
		 Chord               Chord
		 Copy Rect           None. Use 'Draw' instead, and draw what you want
											to see.
         Draw                Draw
		 DrawFocusRect       None. The use of pen styles other than psSolid or
											psClear is not yet catered for.
         Ellipse             Ellipse
		 FillRect            Rectangle (set Pen->Style = bsClear and
											PenWidth = 0 first)
         FloodFill           None
		 FrameRect           Rectangle (set Brush->Style = bsClear and
											PenWidth = 0 first)
         LineTo              LineTo
		 Lock                N/A
         MoveTo              MoveTo
		 Pie                 Pie
		 PolyBezier          PolyBezier
		 PolyBezierTo        PolyBezier, according to the C++ Builder
											documentation, this is functionally
											identical to PolyBezier, so use
											PolyBezier.
         Polygon             Polygon
		 Polyline            PolyLine  (note the capital letter, I hated the
											lack of it in the TCanvas as it was
											inconsistent with AngleArc, ArcTo,
											PolyBezier, RoundRect. However for
											ease in copying code the
											alternative without the capital is
											also offered.)
         Rectangle           Rectangle (note, the rectangle fits exactly inside
											the space defined for it. The 
											boundary line, however thick (so 
											long as a thickness less than the 
											rectangle dimension is chosen), 
											should never exceed it. This is
											not identical behaviour to 
											TCanvas->Rectangle() which doesn't 
											always go exactly up to the bottom 
											right corner, often leaving 1 pixel 
											uncovered.)
		 Refresh             N/A
		 RoundRect           RoundRect
		 StretchDraw         Draw. All bitmaps are drawn stretched.
         TCanvas             TPDFObject
         TextExtent          None, use TextHeight & TextWidth
		 TextHeight          TextHeight()   Official height of font
							 TextHeight(AnsiString text)  	Actual height of
															text presented.
		 TextOut             TextOut
		 TextRect            Use TextOut
		 TextWidth           TextWidth
		 TryLock             N/A
		 Unlock              N/A
		 None                AddResourceImage, adds a bitmap to the pdf file so
											it can be used later
		 None                UseResourceImage, uses a bitmap image on a page
											provided it has already been defined
											in the pdf file using Draw() or
											AddResourceImage() and the object
											number was retained. Bitmaps can
											then be referenced thousands of
											times in a pdf file and there only
											need be one copy of the bitmap.
		 None                URL            See below.
		 None                PageLink       See below.
		 None                HotSpot        See below.
		 None                Comment        Adds a user readable comment to the
											Page Buffer

    The following Properties and their equivalents are present:
        TCanvas             TPDFObject
         Width               PageWidth
         Height              PageHeight
		 Font->Name          Font.Name
		 Font->Color         Font.Color
         Font->Style         FontStyle
		 Brush->Style        Brush.Style    Note only bsSolid & bsClear are
											catered for.
		 Brush->Color        Brush.Color
		 Pen                 Pen            Note only psSolid & psClear pen
											styles are catered for.
		 Pen->Width          PenWidth       Note: PenWidth=0.0 is equivalent to
											 'Minimum Pen Width' not
											 Pen->Style = psClear. It will be
											 different on every device.
		 Pen->Color          PenColour      I am English. I spell Colour with a 'u'. 
		 None                LaunchOnExit,  ShellExecute() opens the PDF File in
											 the PDFObject->BeforeDestruction()
											 event provided the PDF file was
											 created successfully.
											 Defaults to 'true'

    URL: Embeds a hyperlink to a particular URL, and prints some text at the
    same time. The Hyperlink actually defines a clickable area
	
	PageLink: Allows a jump to a set page, similarly to a URL hyperlink.
	
	HotSpot: should be a linkable part of a page, over a graphic etc. It does
	seem to work erratically up to once a page. Beyond that it fails. It should
	work, and I can find no reason for it not to. Probably better not to use.
	
	Comment: Will be invisible in document, just to a reader looking at the
	uncompressed PDF file as a line beginning with '% '. Comments are 
	automatically added to denote the start many of the other functions. This 
	is done to aide de-bugging of an application that uses the PDFObject to
	create PDF files, and echoes the method call into the page buffer.

	EXAMPLE OF HOW TO USE THIS UNIT:
	//creates a 3 page pdf file using many of the capabilities
	//this code goes into the calling unit.
	//requires two image files in .bmp format called "TestBitmap1.bmp" and
	//"TestBitmap2.bmp". These images may be any size, but are ideally square.
	//--------------------------------------------------------------------------
	#include "PageSizeHelp.h"
	#include "PDFUnit.h"
	//--------------------------------------------------------------------------
	TPDFObject *PDFObject = new TPDFObject(this);
	PDFObject->Title = "Test file " + Edit1->Text;
	PDFObject->Author = "I am the Author";
	PDFObject->Creator = "CreatePDF Test Program";
	PDFObject->PageSize = psA4;
	try {
		if (PDFObject->BeginDocument("c:\\temp\\test.pdf")) {
			Graphics::TBitmap *Bitmap = new Graphics::TBitmap();
			Bitmap->LoadFromFile("TestBitmap1.bmp");
			int objno1 = PDFObject->AddResourceImage(Bitmap);
			Bitmap->LoadFromFile("TestBitmap2.bmp");
			int objno2 = PDFObject->AddResourceImage(Bitmap);
			delete Bitmap;
			PDFObject->NewPage();
			PDFObject->PenColour = clRed;
			PDFObject->MoveTo(120.0,15.0);
			PDFObject->LineTo(20.0,195.0);
			PDFObject->PenColour = clBlack;
			PDFObject->PenWidth = 0.5;
			PDFObject->BrushColour = clSilver;
			PDFObject->Rectangle(100,40,195,90);
			PDFObject->BrushColour = clAqua;
			PDFObject->Ellipse(100,40,195,90);
			PDFObject->Brush.Style = bsClear;
			PDFObject->Rectangle(10,100,95,150);
			coordinates Points[4];
			Points[0].X = 125.0;
			Points[0].Y = 125.0;
			Points[1].X = 130.0;
			Points[1].Y = 140.0;
			Points[2].X = 140.0;
			Points[2].Y = 120.0;
			Points[3].X = 150.0;
			Points[3].Y = 150.0;
			PDFObject->PolyBezier(Points,4);
			PDFObject->UseResourceImage(objno1,40,40,50,50);
			PDFObject->UseResourceImage(objno2,50,40,60,50);
			PDFObject->UseResourceImage(objno2,40,50,50,60);
			PDFObject->UseResourceImage(objno1,50,50,60,60);
			//now to stretch the bitmap
			PDFObject->UseResourceImage(objno1,10,60,95,65);
			PDFObject->Font.Color = clBlack;
			PDFObject->Font.Name = "Times";
			PDFObject->FontStyle << fsBold << fsItalic;
			PDFObject->Font.Size = 10;
			PDFObject->TextOut("Hello",100.0,150.0);
			PDFObject->PageLink("Page3",3,55.0,55.0);
			PDFObject->NewPage();
			PDFObject->Font.Name = "Courier";
			PDFObject->Font.Size = 20;
			PDFObject->TextOut("H2",100.0,170.0);
			PDFObject->Font.Name = "Arial";
			PDFObject->TextOut("H3",0,170.0);
			PDFObject->URL("Chestnut Pens Web Site","http://www.chestnutpens.co.uk",100.0,180.0);
			PDFObject->PageLink("Page 1",1,100.0,190.0);
			PDFObject->PageLink("Page 3",3,100.0,200.0);
			PDFObject->EndText();
			PDFObject->NewPage();
			PDFObject->PageSize = psA3;
			PDFObject->PageOrientation = pdLandscape;
			PDFObject->LeftMargin = 0.0;
			PDFObject->TopMargin = 0.0;
			PDFObject->RightMargin = 0.0;
			PDFObject->BottomMargin = 0.0;
			PDFObject->FontStyle << fsUnderline <<fsStrikeOut;
			PDFObject->TextOut("This is Page 3 and it's A3 Landscape, and this should be underlined & struck out",10.0,10.0);
			PDFObject->PenWidth = 1.5;
			PDFObject->BrushColour = clSilver;
			PDFObject->Brush.Style = bsSolid;
			PDFObject->RoundRect(10,40,95,90,10,10);
			PDFObject->AngleArc(250.0,100.0,50.0,30.0,270.0);
			PDFObject->Arc(100.0,100.0,150.0,125.0,150.0,100.0,150.0,125.0);
			PDFObject->Chord(100.0,150.0,150.0,175.0,150.0,150.0,150.0,175.0);
			PDFObject->BrushColour = clTeal;
			PDFObject->Chord(100.0,150.0,150.0,175.0,150.0,175.0,150.0,150.0);
			PDFObject->BrushColour = clSilver;
			PDFObject->Pie(100.0,200.0,150.0,225.0,150.0,200.0,150.0,225.0);
			PDFObject->BrushColour = clNavy;
			PDFObject->Pie(125.0,200.0,175.0,225.0,175.0,225.0,175.0,200.0);

			PDFObject->EndDocument();
			} else {
			String message = "Sorry, the PDF file could not be created";
			String title = "Error";
			Application->MessageBox(message.c_str(),title.c_str(),IDOK);
			}
		}
	__finally {
		delete PDFObject;
		}

	//--------------------------------------------------------------------------

	OTHER NOTES
	This unit used to work with BCB4 Pro, but I've had to change the '#include's
	to get it to work with RAD Studio 10.1 Berlin for C++ Builder. It is a VCL
	using component and care will need to be exercised when using with
	Firemonkey applications. To get it to work with any other C++ Builder
	version should involve no more than changing the '#include's to match your
	version.

	The unit that calls the PDF unit will need the following includes on the
	source or header page:
		#include "PageSizeHelp.h"
		#include "PDFUnit.h"

	You will need to add the following units to the project:
		"PDFUnit.cpp"
		"StringHelp.cpp"
		"GraphicsHelp.cpp"
		"MathsHelp.cpp"


	Some things I really do not understand (well, there are millions of others,
	but only these are related to the PDFUnit):
	1)
	Acrobat Reader always asks if you wish to save changes made to a PDF made
	with this unit. I think this is because there is no compression in the PDF
	file and Acrobat Reader wants to save it with some - I'm not really sure.
	I do know, however, Acrobat Reader uses a more verbose format which does
	not compress so well when stuffed into a .zip file. This does not happen
	with any other reader I have used.
	2)
	Once in a while, when doing a debug run hooked into RAD Studio, it can hang
	and freeze RAD Studio with it. I have never had a program hang when not
	attached to RAD Studio, and using the .exe file that froze when hooked up to
	RAD Studio, it behaves properly when run on its own without RAD Studio.
	When the IDE hangs, the whole lot needs to be killed using 'Task Manager,
	End Process Tree' on RAD Studio. It may be a RAD Studio issue, as I have
	seen reports of this happening in debug mode for other people with their
    programs.


	IN CONCLUSION (positives & negatives):
	1)  +Easy to use
	2)  +Easy to understand
	3)  -No compression
	4)  -No font embedding
	5)  -Fill & Line Style only solid or clear.
	6)  -Only use Pre-defined fonts.
    7)  +No Delphi code to get your head around.
	8)  +Files produced compress really well in zip utilities.

	Beyond that, if there are questions, please contact me at
	    richardandtracy_at_yahoo.com
	Please put something sensible down in the subject line, else it'll get
	deleted automatically.

	Hope this unit is of use. If you improve it, find a way to embed fonts or
	compress the content, please, please share your work and/or send me a copy.

	Regards,

	Richard Williams
*/
//---------------------------------------------------------------------------

__fastcall TPDFObject::TPDFObject(TComponent* Owner)
    : TObject()
{
    PDFOwner = Owner;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::AfterConstruction(void)
{
    //the PDFObject 'OnCreate' Event
    fPageSize = psA4;
    fPageWidth = 210.0;
    fPageHeight = 297.0;
    fPageOrientation = pdPortrait;
    fFileOpen = false;
    fFileName = ExtractFileDir(Application->ExeName) + "\\default.pdf";
    fLeftMargin = 0.0;
    fTopMargin = 0.0;
    fRightMargin = 0.0;
    fBottomMargin = 0.0;
    fPenWidth = 0.25;
    fPenPosX = 0.0;
    fPenPosY = 0.0;
    fPenColour = clBlack;
    fPageOpen = false;
	fTextOpen = false;
	fLaunchOnExit = true;
    fBrushColour = clWhite;
    nextobjectnumber = 5;
    nextpagenumber = 1;
    currentpageparent = 3;

	//Font = new TFont();
	Font.Name = "Arial";
	Font.Size = 10;
    Font.Color = clBlack;
    Pen = new TPen();
	//Brush = new TBrush();
	Brush.Style = bsSolid;
	Brush.Color = clWhite;
    PageBuffer = new TStringList();

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::BeforeDestruction(void)
{
    //the PDFObject 'OnDestroy' Event
    //delete Font;
    delete Pen;
    //delete Brush;
    delete PageBuffer;

    if (FileOpen) {
        EndDocument();
		}

	if (fLaunchOnExit && FileExists(FileName,true)) {
		String Address = FileName;
		String style = "open";
        ShellExecute(Application->Handle,style.c_str(),Address.c_str(),Address.c_str(),NULL,SW_SHOW);
		}

    ObjectList.Length = 0;
    for (int i = 0; i < PageList.Length; i++) {
        PageList[i].FontResource.Length = 0;
        PageList[i].ImageResource.Length = 0;
        PageList[i].AnnotsResource.Length = 0;
        }
    PageList.Length = 0;
    ImageResourceList.Length = 0;
    FontResourceList.Length = 0;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetPageSize(TPageSize Size)
{
	pagesize size = GetPageSize(Size,fPageOrientation);
	fPageWidth = size.PageWidth;
	fPageHeight = size.PageHeight;
	fPageSize = size.PageSize;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetPageWidth(double Width)
{
    fPageWidth = Width;
    fPageSize = psCustom;
    if (fPageWidth > fPageHeight) {
        fPageOrientation = pdLandscape;
        } else {
        fPageOrientation = pdPortrait;
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetPageHeight(double Height)
{
    fPageHeight = Height;
    fPageSize = psCustom;
    if (fPageWidth > fPageHeight) {
        fPageOrientation = pdLandscape;
        } else {
        fPageOrientation = pdPortrait;
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetPageOrientation(TPageDirection Orientation)
{
    fPageOrientation = Orientation;
    if (fPageWidth > fPageHeight && fPageOrientation == /*po*/pdPortrait) {
        double swap = fPageWidth;
        fPageWidth = fPageHeight;
        fPageHeight = swap;
        } else {
        if (fPageWidth < fPageHeight && fPageOrientation == /*po*/pdLandscape) {
            double swap = fPageWidth;
            fPageWidth = fPageHeight;
            fPageHeight = swap;
            }
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetFileName(AnsiString Name)
{
    if (!FileOpen) {
        fFileName = Name;
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetPenColour(TColor Colour)
{
    fPenColour = Colour;
    Pen->Color = Colour;
    if (PageOpen) {
        PageBuffer->Add(GetPenColour());
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetBrushColour(TColor Colour)
{
    fBrushColour = Colour;
	Brush.Color = Colour;
	if (PageOpen) {
		PageBuffer->Add(GetBrushColour());
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::SetPenWidth(double Width)
{
    fPenWidth = Width;
    if (PageOpen) {
		PageBuffer->Add(DP(2,2.8346 * Width) + " w ");
        }
}
//---------------------------------------------------------------------------

bool __fastcall TPDFObject::BeginDocument(AnsiString Name)
{
    if ((output = fopen(Name.c_str(), "wb")) == NULL) {
        return false;
        }

    FileName = Name;
    FileOpen = true;
    streampointer = 0;
    ObjectList.Length = 0;
    nextobjectnumber = 5;
    nextpagenumber = 1;
    currentpageparent = 3;

    //write first characters
    //this will always be object number 0, so don't need to keep its number.
    AddObject(0,65535, true);
    WriteLine("%PDF-1.3");
    WriteLine("%");
    PageOpen = false;

    return true;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::AddObject(int objectnumber, int generation, bool f)
{
    if (objectnumber > ObjectList.Length - 1) {
        int pointer = ObjectList.Length;
        ObjectList.Length = objectnumber + 1;
        while (pointer < ObjectList.Length) {
            ObjectList[pointer].startoffset = 0;
            ObjectList[pointer].generation = 0;
            ObjectList[pointer].f = false;
            pointer++;
            }
        }
    ObjectList[objectnumber].startoffset = streampointer;
    ObjectList[objectnumber].generation = generation;
    ObjectList[objectnumber].f = f;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::NewPage(void)
{
    //starts a new page
    if (PageOpen) FlushPageBuffer();
    //new entry in the PageList
    MakeEnoughPages(nextpagenumber);
    int pointer = nextpagenumber - 1;
    PageList[pointer].pagenumber = nextpagenumber;
    nextpagenumber++;
    if (PageList[pointer].objectnumber == 0) {
        PageList[pointer].objectnumber = nextobjectnumber;
        nextobjectnumber++;
        }
    PageList[pointer].generation = 0;
    //a variable is used here so that I can add parent ability at a later date.
    PageList[pointer].parentobjectnumber = currentpageparent;
    PageList[pointer].parentgeneration = 0;
    PageList[pointer].contentsobjectnumber = 0;
    PageList[pointer].contentsgeneration = 0;
    PageList[pointer].resourceImageB = false;
    PageList[pointer].resourceImageC = false;
    PageList[pointer].resourceText = false;
    PageList[pointer].FontResource.Length = 0;
    PageList[pointer].ImageResource.Length = 0;
    PageList[pointer].used = true;
    PageOpen = true;
    //set up colours
    PageBuffer->Add(GetPenColour());
    PageBuffer->Add(GetBrushColour());
    //line cap style round
    PageBuffer->Add("1 J ");
    //line join style Mitred
    PageBuffer->Add("0 j ");
    //PageBuffer->Add("q ");
    //PageBuffer->Add("0 0 0 0 1 1 cm ");
    PenWidth = PenWidth;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::EndDocument(void)
{
    //finishes the PDF document and clears up.
    if (PageOpen) {
        if(TextOpen) EndText();
        FlushPageBuffer();
        }
	WriteCatalog();
	WriteOutlines();
	WritePages();
    WriteProducer();
    int xrefstart = streampointer;
    WriteLine("xref");
    WriteLine("0 " + AnsiString(ObjectList.Length) + " ");
	AnsiString text;
	for (int i = 0; i < ObjectList.Length; i++) {
		if (i > 0 && ObjectList[i].startoffset == 0) {
			ObjectList[i].f = true;
            ObjectList[i].startoffset = ObjectList[i-1].startoffset;
			}
		text =  FixedLenZeroPadded(10,ObjectList[i].startoffset) + " " + FixedLenZeroPadded(5,ObjectList[i].generation) + " ";
		if (ObjectList[i].f) {
            text = text + "f ";
            } else {
            text = text + "n ";
            }
        WriteLine(text);
        }
    WriteLine("trailer");
    WriteLine("<< /Size " + AnsiString(ObjectList.Length) + " ");
    WriteLine("/Root 1 0 R ");
    WriteLine("/Info 4 0 R ");
    WriteLine(">>");
    WriteLine("startxref");
    WriteLine(AnsiString(xrefstart));
    WriteLine("%%EOF");
    fclose(output);
    FileOpen = false;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::FlushPageBuffer(void)
{
    //Flushes the page buffer into the pdf file.
    //The page being written is the last in the page list
    //first write the contents object
    WritePageContents();
    //Then the page object
    WritePage();
    PageOpen = false;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WriteCatalog(void)
{
    //writes the standard catalog object
    AddObject(1,0,false);
    WriteLine("1 0 obj ");
    WriteLine("<< /Type /Catalog ");
    WriteLine("/Outlines 2 0 R ");
    WriteLine("/Pages 3 0 R ");
    WriteLine(">> ");
    WriteLine("endobj ");
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WriteOutlines(void)
{
    //writes the standard outlines object
    AddObject(2,0,false);
    WriteLine("2 0 obj ");
    WriteLine("<< /Type /Outlines ");
    WriteLine("/Count 0 ");
    WriteLine(">> ");
    WriteLine("endobj ");
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WritePageContents(void)
{
    //creates an object that represents the contents of the current pagebuffer
    if (TextOpen) EndText();
    AddObject(nextobjectnumber,0,false);
    //int pointer = PageList.Length - 1;
    int pointer = nextpagenumber - 2;
    PageList[pointer].contentsobjectnumber = nextobjectnumber;
    PageList[pointer].contentsgeneration = 0;
    WriteLine(AnsiString(PageList[pointer].contentsobjectnumber) + " " + AnsiString(PageList[pointer].contentsgeneration) + " obj ");
    WriteLine("<< /Length " + AnsiString(nextobjectnumber + 1) + " 0 R >>");
    WriteLine("stream");
    int start = streampointer;
    //PageBuffer->Add("Q ");
    for (int i = 0; i < PageBuffer->Count; i++) {
        WriteLine(PageBuffer->Strings[i]);
        }
    int end = streampointer;
    WriteLine("endstream");
    WriteLine("endobj ");

    PageBuffer->Clear();

    nextobjectnumber++;
    AddObject(nextobjectnumber,0,false);
    WriteLine(AnsiString(nextobjectnumber) + " 0 obj ");
    //WriteLine("/Length " + AnsiString(end - start) + " ");
    WriteLine(AnsiString(end - start) + " ");
    WriteLine("endobj ");
    nextobjectnumber++;


}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WritePage(void)
{
    //writes the current page object, is called by 'FlushPageBuffer()'
    //int pointer = PageList.Length - 1;
    int pointer = nextpagenumber - 2;
    AddObject(PageList[pointer].objectnumber,PageList[pointer].generation,false);
    WriteLine(AnsiString(PageList[pointer].objectnumber) + " " + AnsiString(PageList[pointer].generation) + " obj ");
    WriteLine("<< /Type /Page ");
	//Page Size
	WriteLine("/MediaBox [0 0 " + DP(2,PageWidth*2.8346) + " " + DP(2,PageHeight*2.8346) + " ] ");
	//Margins
	if (LeftMargin != 0.0 || TopMargin != 0.0 || RightMargin != 0.0 || BottomMargin != 0.0) {
		WriteLine("/CropBox [" + DP(2,LeftMargin*2.8346) + " " + DP(2,BottomMargin*2.8346) + " " + DP(2,(PageWidth-RightMargin)*2.8346) + " " + DP(2,(PageHeight-TopMargin)*2.8346) + " ] ");
		}
	WriteLine("/Rotate 0 ");
	WriteLine("/PZ 1.0");
    WriteLine("/Parent " + AnsiString(PageList[pointer].parentobjectnumber) + " " + AnsiString(PageList[pointer].parentgeneration) + " R ");
    if (PageList[pointer].AnnotsResource.Length) {
        WriteLine("/Annots [");
        for (int i = 0; i < PageList[pointer].AnnotsResource.Length; i++) {
            WriteLine(AnsiString(PageList[pointer].AnnotsResource[i].objectnumber) + " " + AnsiString(PageList[pointer].AnnotsResource[i].generation) + " R ");
            }
        WriteLine("] ");
        }
    WriteLine("/Resources << ");
    WriteLine("/ProcSet [ ");
    WriteLine("/PDF ");
    if (PageList[pointer].resourceImageB) WriteLine("/ImageB ");
    if (PageList[pointer].resourceImageC) WriteLine("/ImageC ");
    if (PageList[pointer].resourceText) WriteLine("/Text ");
    WriteLine("] ");
    if (PageList[pointer].FontResource.Length == 1) {
        int pointer2 = 0;
        bool found = false;
        while (!found && pointer2 < FontResourceList.Length) {
            if (PageList[pointer].FontResource[0].objectnumber == FontResourceList[pointer2].objectnumber) {
                found = true;
                pointer2--;
                }
            pointer2++;
            }
        if (found) WriteLine("/Font <</" + FontResourceList[pointer2].Name + " " + AnsiString(PageList[pointer].FontResource[0].objectnumber) + " " + AnsiString(PageList[pointer].FontResource[0].generation) + " R >> ");
        } else {
        if (PageList[pointer].FontResource.Length > 1) {
            WriteLine("/Font << ");
            for (int i = 0; i < PageList[pointer].FontResource.Length; i++) {
                int pointer2 = 0;
                bool found = false;
                while (!found && pointer2 < ImageResourceList.Length) {
                    if (PageList[pointer].FontResource[i].objectnumber == FontResourceList[pointer2].objectnumber) {
                        found = true;
                        pointer2--;
                        }
                    pointer2++;
                    }
                if (found) WriteLine("/" + FontResourceList[pointer2].Name + " " + AnsiString(PageList[pointer].FontResource[i].objectnumber) + " " + AnsiString(PageList[pointer].FontResource[i].generation) + " R ");
                }
            WriteLine(">> ");
            }
        }
    if (PageList[pointer].ImageResource.Length == 1) {
        int pointer2 = 0;
        bool found = false;
        while (!found && pointer2 < ImageResourceList.Length) {
            if (PageList[pointer].ImageResource[0].objectnumber == ImageResourceList[pointer2].objectnumber) {
                found = true;
                pointer2--;
                }
            pointer2++;
            }
        if (found) WriteLine("/XObject <</" + ImageResourceList[pointer2].Name + " " + AnsiString(PageList[pointer].ImageResource[0].objectnumber) + " " + AnsiString(PageList[pointer].ImageResource[0].generation) + " R >> ");
        } else {
        if (PageList[pointer].ImageResource.Length > 1) {
            WriteLine("/XObject << ");
            for (int i = 0; i < PageList[pointer].ImageResource.Length; i++) {
                int pointer2 = 0;
                bool found = false;
                while (!found && pointer2 < ImageResourceList.Length) {
                    if (PageList[pointer].ImageResource[i].objectnumber == ImageResourceList[pointer2].objectnumber) {
                        found = true;
                        pointer2--;
                        }
                    pointer2++;
                    }
                if (found) WriteLine("/" + ImageResourceList[pointer2].Name + " " + AnsiString(PageList[pointer].ImageResource[i].objectnumber) + " " + AnsiString(PageList[pointer].ImageResource[i].generation) + " R ");
                }
            WriteLine(">> ");
            }
        }
    WriteLine(">> ");
    WriteLine("/Contents " + AnsiString(PageList[pointer].contentsobjectnumber) + " " + AnsiString(PageList[pointer].contentsgeneration) + " R ");
    WriteLine(">> ");
    WriteLine("endobj ");
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WritePages(void)
{
    //writes the standard pages object
    AddObject(3,0,false);
    WriteLine("3 0 obj ");
    WriteLine("<< /Type /Pages ");
    if (PageList.Length) {
        WriteLine("/Kids [ ");
        for (int i = 0; i < PageList.Length; i++) {
            WriteLine(AnsiString(PageList[i].objectnumber) + " " + AnsiString(ObjectList[PageList[i].objectnumber].generation) + " R ");
            }
        WriteLine("] ");
        }
    WriteLine("/Count "+ AnsiString(PageList.Length) + " ");
    WriteLine(">> ");
    WriteLine("endobj ");
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WriteProducer(void)
{
    //writes the standard producer object
    AddObject(4,0,false);
	WriteLine("4 0 obj ");
	//It would be appreciated if you do not change this line - but you are in charge of the source code!
    WriteLine("<< /Producer (Chestnut Pens free PDFObject for C++ Builder) ");
    if (Title.Length()) {
		WriteLine("/Title (" + SanitiseText(Title) + ") ");
        }
    if (Author.Length()) {
		WriteLine("/Author (" + SanitiseText(Author) + ") ");
        }
    if (Creator.Length()) {
        WriteLine("/Creator (" + SanitiseText(Creator) + ") ");
		}
	if (Subject.Length()) {
		WriteLine("/Subject (" + SanitiseText(Subject) + ") ");
		}
	if (Keywords.Length()) {
		WriteLine("/Keywords (" + SanitiseText(Keywords) + ") ");
		}
	TDateTime Current;
    Current = Current.CurrentDateTime();
    AnsiString text = "D:" + Current.FormatString("yyyymmddhhnnss") + TimeZone();
    WriteLine("/CreationDate (" + text + ") ");
    WriteLine("/ModDate (" + text + ") ");
    WriteLine(">> ");
    WriteLine("endobj ");
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WriteLine(AnsiString text)
{
    if (FileOpen) {
        fwrite(text.c_str(),strlen(text.c_str()),1,output);
        streampointer += strlen(text.c_str());
        CRLF();
        }

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::WritePageLine(AnsiString text)
{
    if (FileOpen) {
        fwrite(text.c_str(),strlen(text.c_str()),1,output);
        streampointer += strlen(text.c_str());
        //CRLF();
        }

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::CRLF(void)
{
    //carriage return, line feed.
    if (FileOpen) {
        char text = 13;
        fwrite(&text,1,1,output);
        text = 10;
        fwrite(&text,1,1,output);
        streampointer += 2;
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::AngleArc(double CentreX, double CentreY, double Radius, double StartAngle, double SweepAngle)
{
	//reproduces the TCanvas->AngleArc method on the PDF page
	//it uses the MathsHelp unit for the calculation of the Bezier points to draw the arc,
	//so must operate with  +ve Y = Up co-ordinate system, not the one usually used
	//as a result of this, everything is calculated in units of 'point' (1/72") from the bottom left corner of the page

	//first convert to PDF co-ordinates
	vector2D centre;
	double localradius, localpenposX, localpenposY;

	centre.x = CentreX * 2.8346;
	centre.y = (PageHeight - CentreY) * 2.8346;
	localradius = Radius * 2.8346;
	localpenposX = PenPosX * 2.8346;
	localpenposY = (PageHeight - PenPosY) * 2.8346;

	//OK, now to find how many arcs to create (up to 90 degrees per arc)
	DynamicArray <vector2D> BezierPoints;
	vector2D Points[4];
	int numberofarcs = ((int)(fabs(SweepAngle)/90.0)) + 1;
	BezierPoints.Length = numberofarcs * 3 + 1;
	if (SweepAngle < 0.0) {
		StartAngle = StartAngle + SweepAngle;
		SweepAngle = SweepAngle * -1.0;
		}
	double sweepperarc = SweepAngle/((double)(numberofarcs));
	bool problemflag = false;

	for (int i = 0; i < numberofarcs; i++) {
		double startangle = StartAngle + ((double)(i)) * sweepperarc;
		double endangle = StartAngle + ((double)(i + 1)) * sweepperarc;
		bool result = BezierArc2D(Points,centre,localradius,startangle,endangle);
		if (result) {
			if (i == 0) {
				BezierPoints[0] = Points[0];
				}
			BezierPoints[3 * i + 1] = Points[1];
			BezierPoints[3 * i + 2] = Points[2];
			BezierPoints[3 * i + 3] = Points[3];
			} else {
			problemflag = true;
            }
		}

	//That's all the necessary points calculated, now to do something with them
	if (!problemflag) {
		if (!PageOpen) NewPage();
		if (TextOpen) EndText();
		PageBuffer->Add("% AngleArc(" + AnsiString(CentreX) + "," + AnsiString(CentreY) + "," + AnsiString(Radius) + "," + AnsiString(StartAngle) + "," + AnsiString(SweepAngle) + ")"  );
		PageBuffer->Add(DP(3,localpenposX) + " " + DP(3,localpenposY) + " m ");
		if (localpenposX != BezierPoints[0].x || localpenposY != BezierPoints[0].y) {
			PageBuffer->Add(DP(3,BezierPoints[0].x) + " " + DP(3,BezierPoints[0].y) + " l ");
			}
		for (int i = 0;  i < numberofarcs; i++) {
			PageBuffer->Add(DP(3,BezierPoints[3 * i + 1].x) + " " + DP(3,BezierPoints[3 * i + 1].y) + " " + DP(3,BezierPoints[3 * i + 2].x)  + " " + DP(3,BezierPoints[3 * i + 2].y) + " " + DP(3,BezierPoints[3 * i + 3].x)  + " " + DP(3,BezierPoints[3 * i + 3].y) + " c ");
			}
		PageBuffer->Add("S ");
		localpenposX = BezierPoints[BezierPoints.Length - 1].x;
		localpenposY = BezierPoints[BezierPoints.Length - 1].y;
		}

	BezierPoints.Length = 0;

	//just before finishing, get the new pen position in mm from the top left corner of the page
	PenPosX = localpenposX/2.8346;
	PenPosY = PageHeight - localpenposY/2.8346;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Arc(double X1, double Y1, double X2, double Y2, double X3, double Y3, double X4, double Y4)
{
	//Reproduces the TCanvas->Arc method on the PDF page
	//It uses the MathsHelp unit for the calculation of the Bezier points to draw the arc,
	//so must operate with  +ve Y = Up co-ordinate system, not the one usually used.
	//As a result of this, everything is calculated in units of 'point' (1/72") from the bottom left corner of the page.
	//The ellipse is first scaled vertically to turn it into a circle, the segments are calculated as per 'AngleArc' then scaled back into an ellipse

	//first convert to PDF co-ordinates
	vector2D topleft,bottomright,startpoint, endpoint;
	topleft.x = X1 * 2.8346;
	topleft.y = (PageHeight - Y1) * 2.8346;
	bottomright.x = X2 * 2.8346;
	bottomright.y = (PageHeight - Y2) * 2.8346;
	startpoint.x = X3 * 2.8346;
	startpoint.y = (PageHeight - Y3) * 2.8346;
	endpoint.x = X4 * 2.8346;
	endpoint.y = (PageHeight - Y4) * 2.8346;
	//check to see the ellipse had a dimension in both X and Y
	if ((bottomright.x-topleft.x) > 0.0 && (topleft.y - bottomright.y) > 0.0) {

		double yscale = (topleft.y - bottomright.y)/(bottomright.x-topleft.x);
		vector2D centre, startpointtrans, endpointtrans;
		double localradius;

		centre.x = (bottomright.x + topleft.x)/2.0;
		centre.y = (topleft.y + bottomright.y)/2.0;
		localradius = bottomright.x - centre.x;

		//vertical stretch/shrink of ellipse start/end points to make into a circle
		startpointtrans.x = startpoint.x;
		startpointtrans.y =  centre.y + (startpoint.y-centre.y)/yscale;
		endpointtrans.x = endpoint.x;
		endpointtrans.y =  centre.y + (endpoint.y-centre.y)/yscale;
		double StartAngle, EndAngle, SweepAngle;
		startpointtrans = UnitVector(SubtractVector(startpointtrans,centre));
		endpointtrans = UnitVector(SubtractVector(endpointtrans,centre));
		StartAngle = acosD(startpointtrans.x);
		if (startpointtrans.y < 0.0) StartAngle = 360.0 - StartAngle;
		if (startpointtrans.x == -1.0) StartAngle = 180.0;
		EndAngle = acosD(endpointtrans.x);
		if (endpointtrans.y < 0.0) EndAngle = 360.0 - EndAngle;
		if (endpointtrans.x == -1.0) EndAngle = 180.0;
		while (EndAngle < StartAngle) EndAngle += 360.0;
		SweepAngle = EndAngle - StartAngle;

		//OK, now to find how many arcs to create (up to 90 degrees per arc)
		DynamicArray <vector2D> BezierPoints;
		vector2D Points[4];
		int numberofarcs = ((int)(SweepAngle/90.0)) + 1;
		BezierPoints.Length = numberofarcs * 3 + 1;
		double sweepperarc = SweepAngle/((double)(numberofarcs));
		bool problemflag = false;

		for (int i = 0; i < numberofarcs; i++) {
			double startangle = StartAngle + ((double)(i)) * sweepperarc;
			double endangle = StartAngle + ((double)(i + 1)) * sweepperarc;
			bool result = BezierArc2D(Points,centre,localradius,startangle,endangle);
			if (result) {
				if (i == 0) {
					BezierPoints[0] = Points[0];
					}
				BezierPoints[3 * i + 1] = Points[1];
				BezierPoints[3 * i + 2] = Points[2];
				BezierPoints[3 * i + 3] = Points[3];
				} else {
				problemflag = true;
				}
			}


		//That's all the necessary points calculated, now to do something with them
		if (!problemflag) {
			if (!PageOpen) NewPage();
			if (TextOpen) EndText();
			//That's all the points calculated assuming they are on a circular arc. Now to scale them all in the Y direction
			for (int i = 0; i < BezierPoints.Length; i++) {
				BezierPoints[i].y = centre.y + (BezierPoints[i].y - centre.y) * yscale;
				}
			PageBuffer->Add("% Arc(" + AnsiString(X1) + "," + AnsiString(Y1)+ "," + AnsiString(X2) + "," + AnsiString(Y2) + "," + AnsiString(X3) + "," + AnsiString(Y3)+ "," + AnsiString(X4) + "," + AnsiString(Y4) + ")");
			PageBuffer->Add(DP(3,BezierPoints[0].x) + " " + DP(3,BezierPoints[0].y) + " m ");
			for (int i = 0;  i < numberofarcs; i++) {
				PageBuffer->Add(DP(3,BezierPoints[3 * i + 1].x) + " " + DP(3,BezierPoints[3 * i + 1].y) + " " + DP(3,BezierPoints[3 * i + 2].x)  + " " + DP(3,BezierPoints[3 * i + 2].y) + " " + DP(3,BezierPoints[3 * i + 3].x)  + " " + DP(3,BezierPoints[3 * i + 3].y) + " c ");
				}
			PageBuffer->Add("S ");
			//just before finishing, get the new pen position in mm from the top left corner of the page
			PenPosX = BezierPoints[BezierPoints.Length - 1].x/2.8346;
			PenPosY = PageHeight - BezierPoints[BezierPoints.Length - 1].y/2.8346;
			}

		BezierPoints.Length = 0;


		}
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Chord(double X1, double Y1, double X2, double Y2, double X3, double Y3, double X4, double Y4)
{
	//Reproduces the TCanvas->Chord method on the PDF page
	//It uses the MathsHelp unit for the calculation of the Bezier points to draw the arc,
	//so must operate with  +ve Y = Up co-ordinate system, not the one usually used.
	//As a result of this, everything is calculated in units of 'point' (1/72") from the bottom left corner of the page.
	//The ellipse is first scaled vertically to turn it into a circle, the segments are calculated as per 'AngleArc' then scaled back into an ellipse

	//first convert to PDF co-ordinates
	vector2D topleft,bottomright,startpoint, endpoint;
	topleft.x = X1 * 2.8346;
	topleft.y = (PageHeight - Y1) * 2.8346;
	bottomright.x = X2 * 2.8346;
	bottomright.y = (PageHeight - Y2) * 2.8346;
	startpoint.x = X3 * 2.8346;
	startpoint.y = (PageHeight - Y3) * 2.8346;
	endpoint.x = X4 * 2.8346;
	endpoint.y = (PageHeight - Y4) * 2.8346;
	//check to see the ellipse had a dimension in both X and Y
	if ((bottomright.x-topleft.x) > 0.0 && (topleft.y - bottomright.y) > 0.0) {

		double yscale = (topleft.y - bottomright.y)/(bottomright.x-topleft.x);
		vector2D centre, startpointtrans, endpointtrans;
		double localradius;

		centre.x = (bottomright.x + topleft.x)/2.0;
		centre.y = (topleft.y + bottomright.y)/2.0;
		localradius = bottomright.x - centre.x;

		//vertical stretch/shrink of ellipse start/end points to make into a circle
		startpointtrans.x = startpoint.x;
		startpointtrans.y =  centre.y + (startpoint.y-centre.y)/yscale;
		endpointtrans.x = endpoint.x;
		endpointtrans.y =  centre.y + (endpoint.y-centre.y)/yscale;
		double StartAngle, EndAngle, SweepAngle;
		startpointtrans = UnitVector(SubtractVector(startpointtrans,centre));
		endpointtrans = UnitVector(SubtractVector(endpointtrans,centre));
		StartAngle = acosD(startpointtrans.x);
		if (startpointtrans.y < 0.0) StartAngle = 360.0 - StartAngle;
		if (startpointtrans.x == -1.0) StartAngle = 180.0;
		EndAngle = acosD(endpointtrans.x);
		if (endpointtrans.y < 0.0) EndAngle = 360.0 - EndAngle;
		if (endpointtrans.x == -1.0) EndAngle = 180.0;
		while (EndAngle < StartAngle) EndAngle += 360.0;
		SweepAngle = EndAngle - StartAngle;

		//OK, now to find how many arcs to create (up to 90 degrees per arc)
		DynamicArray <vector2D> BezierPoints;
		vector2D Points[4];
		int numberofarcs = ((int)(SweepAngle/90.0)) + 1;
		BezierPoints.Length = numberofarcs * 3 + 1;
		double sweepperarc = SweepAngle/((double)(numberofarcs));
		bool problemflag = false;

		for (int i = 0; i < numberofarcs; i++) {
			double startangle = StartAngle + ((double)(i)) * sweepperarc;
			double endangle = StartAngle + ((double)(i + 1)) * sweepperarc;
			bool result = BezierArc2D(Points,centre,localradius,startangle,endangle);
			if (result) {
				if (i == 0) {
					BezierPoints[0] = Points[0];
					}
				BezierPoints[3 * i + 1] = Points[1];
				BezierPoints[3 * i + 2] = Points[2];
				BezierPoints[3 * i + 3] = Points[3];
				} else {
				problemflag = true;
				}
			}


		//That's all the necessary points calculated, now to do something with them
		if (!problemflag) {
			if (!PageOpen) NewPage();
			if (TextOpen) EndText();
			//That's all the points calculated assuming they are on a circular arc. Now to scale them all in the Y direction
			for (int i = 0; i < BezierPoints.Length; i++) {
				BezierPoints[i].y = centre.y + (BezierPoints[i].y - centre.y) * yscale;
				}
			PageBuffer->Add("% Chord(" + AnsiString(X1) + "," + AnsiString(Y1)+ "," + AnsiString(X2) + "," + AnsiString(Y2) + "," + AnsiString(X3) + "," + AnsiString(Y3)+ "," + AnsiString(X4) + "," + AnsiString(Y4) + ")");
			PageBuffer->Add(DP(3,BezierPoints[0].x) + " " + DP(3,BezierPoints[0].y) + " m ");
			for (int i = 0;  i < numberofarcs; i++) {
				PageBuffer->Add(DP(3,BezierPoints[3 * i + 1].x) + " " + DP(3,BezierPoints[3 * i + 1].y) + " " + DP(3,BezierPoints[3 * i + 2].x)  + " " + DP(3,BezierPoints[3 * i + 2].y) + " " + DP(3,BezierPoints[3 * i + 3].x)  + " " + DP(3,BezierPoints[3 * i + 3].y) + " c ");
				}
			PageBuffer->Add(DP(3,BezierPoints[0].x) + " " + DP(3,BezierPoints[0].y) + " l ");
			if (Brush.Style == bsSolid) {
				PageBuffer->Add("B ");
				} else {
				PageBuffer->Add("S ");
				}
			//just before finishing, get the new pen position in mm from the top left corner of the page
			PenPosX = BezierPoints[0].x/2.8346;
			PenPosY = PageHeight - BezierPoints[0].y/2.8346;
			}

		BezierPoints.Length = 0;


		}
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Ellipse( double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY)
{
	if (!PageOpen) NewPage();
	if (TextOpen) EndText();
	TColor oldcolour = PenColour;
	if (Pen->Style == psClear) PenColour = BrushColour;
	//adjust size so that outer edge of line is exactly on the defined boundary
	TopLeftX = TopLeftX + PenWidth/2.0;
	TopLeftY = TopLeftY + PenWidth/2.0;
	BottomRightX = BottomRightX - PenWidth/2.0;
	BottomRightY = BottomRightY - PenWidth/2.0;
	DynamicArray <coordinates> Coordinates;
	Coordinates.Length = 12;
	coordinates delta;
	//the value below is half of the magic 'k' for a bezier line. K = 0.5522847498 K=(4/3)*(sqrt(2)-1)
	delta.X = 0.2238576251 * (BottomRightX - TopLeftX);
	delta.Y = 0.2238576251 * (BottomRightY - TopLeftY);
	Coordinates[0].X = TopLeftX;
	Coordinates[0].Y = (BottomRightY + TopLeftY)/2.0;
	Coordinates[1].X = TopLeftX;
	Coordinates[1].Y = BottomRightY - delta.Y;
	Coordinates[2].X = TopLeftX + delta.X;
	Coordinates[2].Y = BottomRightY;
	Coordinates[3].X = (BottomRightX + TopLeftX)/2.0;
	Coordinates[3].Y = BottomRightY;
	Coordinates[4].X = BottomRightX - delta.X;
	Coordinates[4].Y = BottomRightY;
	Coordinates[5].X = BottomRightX;
	Coordinates[5].Y = Coordinates[1].Y;
	Coordinates[6].X = BottomRightX;
	Coordinates[6].Y = Coordinates[0].Y;
	Coordinates[7].X = BottomRightX;
	Coordinates[7].Y = TopLeftY + delta.Y;
	Coordinates[8].X = Coordinates[4].X;
	Coordinates[8].Y = TopLeftY;
    Coordinates[9].X = Coordinates[3].X;
	Coordinates[9].Y = TopLeftY;
	Coordinates[10].X = Coordinates[2].X;
	Coordinates[10].Y = TopLeftY;
    Coordinates[11].X = TopLeftX;
	Coordinates[11].Y = Coordinates[7].Y;
	for (int i = 0; i < Coordinates.Length; i++) {
		Coordinates[i].X = 2.8346 * Coordinates[i].X;
        Coordinates[i].Y = 2.8346 * (PageHeight - Coordinates[i].Y);
		}
	PageBuffer->Add("% Ellipse(" + AnsiString(TopLeftX) + "," + AnsiString(TopLeftY)  + "," + AnsiString(BottomRightX)  + "," + AnsiString(BottomRightY) + ")");
	PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " m ");
	PageBuffer->Add(DP(3,Coordinates[1].X) + " " + DP(3,Coordinates[1].Y) + " " + DP(3,Coordinates[2].X)  + " " + DP(3,Coordinates[2].Y) + " " + DP(3,Coordinates[3].X)  + " " + DP(3,Coordinates[3].Y) + " c ");
    PageBuffer->Add(DP(3,Coordinates[4].X) + " " + DP(3,Coordinates[4].Y) + " " + DP(3,Coordinates[5].X)  + " " + DP(3,Coordinates[5].Y) + " " + DP(3,Coordinates[6].X)  + " " + DP(3,Coordinates[6].Y) + " c ");
	PageBuffer->Add(DP(3,Coordinates[7].X) + " " + DP(3,Coordinates[7].Y) + " " + DP(3,Coordinates[8].X)  + " " + DP(3,Coordinates[8].Y) + " " + DP(3,Coordinates[9].X)  + " " + DP(3,Coordinates[9].Y) + " c ");
	PageBuffer->Add(DP(3,Coordinates[10].X) + " " + DP(3,Coordinates[10].Y) + " " + DP(3,Coordinates[11].X)  + " " + DP(3,Coordinates[11].Y) + " " + DP(3,Coordinates[0].X)  + " " + DP(3,Coordinates[0].Y) + " c ");
	//if (Brush->Style == bsSolid) {
	if (Brush.Style == bsSolid) {
		PageBuffer->Add("B ");
		} else {
		PageBuffer->Add("S ");
		}
	PenPosX = TopLeftX;
	PenPosY = (BottomRightY + TopLeftY)/2.0;
	if (Pen->Style == psClear) PenColour = oldcolour;

	Coordinates.Length = 0;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::LineTo(double X, double Y)
{
    if (!PageOpen) NewPage();
    if (TextOpen) EndText();
	PageBuffer->Add("% LineTo(" + AnsiString(X) + "," + AnsiString(Y) + ")");
	PageBuffer->Add(DP(3,2.8346 * X) + " " + DP(3,2.8346 * (PageHeight - Y)) + " l ");
	PageBuffer->Add("S ");
	PenPosX = X;
    PenPosY = Y;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::MoveTo(double X, double Y)
{
	if (!PageOpen) NewPage();
	if (TextOpen) EndText();
	PageBuffer->Add("% MoveTo(" + AnsiString(X) + "," + AnsiString(Y) + ")");
	PageBuffer->Add(DP(3,2.8346 * X) + " " + DP(3,2.8346 * (PageHeight - Y)) + " m ");
    PenPosX = X;
    PenPosY = Y;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Pie(double X1, double Y1, double X2, double Y2, double X3, double Y3, double X4, double Y4)
{
	//Reproduces the TCanvas->Pie method on the PDF page
	//It uses the MathsHelp unit for the calculation of the Bezier points to draw the arc,
	//so must operate with  +ve Y = Up co-ordinate system, not the one usually used.
	//As a result of this, everything is calculated in units of 'point' (1/72") from the bottom left corner of the page.
	//The ellipse is first scaled vertically to turn it into a circle, the segments are calculated as per 'AngleArc' then scaled back into an ellipse

	//first convert to PDF co-ordinates
	vector2D topleft,bottomright,startpoint, endpoint;
	topleft.x = X1 * 2.8346;
	topleft.y = (PageHeight - Y1) * 2.8346;
	bottomright.x = X2 * 2.8346;
	bottomright.y = (PageHeight - Y2) * 2.8346;
	startpoint.x = X3 * 2.8346;
	startpoint.y = (PageHeight - Y3) * 2.8346;
	endpoint.x = X4 * 2.8346;
	endpoint.y = (PageHeight - Y4) * 2.8346;
	//check to see the ellipse had a dimension in both X and Y
	if ((bottomright.x-topleft.x) > 0.0 && (topleft.y - bottomright.y) > 0.0) {

		double yscale = (topleft.y - bottomright.y)/(bottomright.x-topleft.x);
		vector2D centre, startpointtrans, endpointtrans;
		double localradius;

		centre.x = (bottomright.x + topleft.x)/2.0;
		centre.y = (topleft.y + bottomright.y)/2.0;
		localradius = bottomright.x - centre.x;

		//vertical stretch/shrink of ellipse start/end points to make into a circle
		startpointtrans.x = startpoint.x;
		startpointtrans.y =  centre.y + (startpoint.y-centre.y)/yscale;
		endpointtrans.x = endpoint.x;
		endpointtrans.y =  centre.y + (endpoint.y-centre.y)/yscale;
		double StartAngle, EndAngle, SweepAngle;
		startpointtrans = UnitVector(SubtractVector(startpointtrans,centre));
		endpointtrans = UnitVector(SubtractVector(endpointtrans,centre));
		StartAngle = acosD(startpointtrans.x);
		if (startpointtrans.y < 0.0) StartAngle = 360.0 - StartAngle;
		if (startpointtrans.x == -1.0) StartAngle = 180.0;
		EndAngle = acosD(endpointtrans.x);
		if (endpointtrans.y < 0.0) EndAngle = 360.0 - EndAngle;
		if (endpointtrans.x == -1.0) EndAngle = 180.0;
		while (EndAngle < StartAngle) EndAngle += 360.0;
		SweepAngle = EndAngle - StartAngle;

		//OK, now to find how many arcs to create (up to 90 degrees per arc)
		DynamicArray <vector2D> BezierPoints;
		vector2D Points[4];
		int numberofarcs = ((int)(SweepAngle/90.0)) + 1;
		BezierPoints.Length = numberofarcs * 3 + 1;
		double sweepperarc = SweepAngle/((double)(numberofarcs));
		bool problemflag = false;

		for (int i = 0; i < numberofarcs; i++) {
			double startangle = StartAngle + ((double)(i)) * sweepperarc;
			double endangle = StartAngle + ((double)(i + 1)) * sweepperarc;
			bool result = BezierArc2D(Points,centre,localradius,startangle,endangle);
			if (result) {
				if (i == 0) {
					BezierPoints[0] = Points[0];
					}
				BezierPoints[3 * i + 1] = Points[1];
				BezierPoints[3 * i + 2] = Points[2];
				BezierPoints[3 * i + 3] = Points[3];
				} else {
				problemflag = true;
				}
			}


		//That's all the necessary points calculated, now to do something with them
		if (!problemflag) {
			if (!PageOpen) NewPage();
			if (TextOpen) EndText();
			//That's all the points calculated assuming they are on a circular arc. Now to scale them all in the Y direction
			for (int i = 0; i < BezierPoints.Length; i++) {
				BezierPoints[i].y = centre.y + (BezierPoints[i].y - centre.y) * yscale;
				}
			PageBuffer->Add("% Pie(" + AnsiString(X1) + "," + AnsiString(Y1)+ "," + AnsiString(X2) + "," + AnsiString(Y2) + "," + AnsiString(X3) + "," + AnsiString(Y3)+ "," + AnsiString(X4) + "," + AnsiString(Y4) + ")");
			PageBuffer->Add(DP(3,BezierPoints[0].x) + " " + DP(3,BezierPoints[0].y) + " m ");
			for (int i = 0;  i < numberofarcs; i++) {
				PageBuffer->Add(DP(3,BezierPoints[3 * i + 1].x) + " " + DP(3,BezierPoints[3 * i + 1].y) + " " + DP(3,BezierPoints[3 * i + 2].x)  + " " + DP(3,BezierPoints[3 * i + 2].y) + " " + DP(3,BezierPoints[3 * i + 3].x)  + " " + DP(3,BezierPoints[3 * i + 3].y) + " c ");
				}
			PageBuffer->Add(DP(3,centre.x) + " " + DP(3,centre.y) + " l ");
			PageBuffer->Add(DP(3,BezierPoints[0].x) + " " + DP(3,BezierPoints[0].y) + " l ");
			if (Brush.Style == bsSolid) {
				PageBuffer->Add("B ");
				} else {
				PageBuffer->Add("S ");
				}
			//just before finishing, get the new pen position in mm from the top left corner of the page
			PenPosX = BezierPoints[0].x/2.8346;
			PenPosY = PageHeight - BezierPoints[0].y/2.8346;
			}

		BezierPoints.Length = 0;


		}
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::PolyBezier(coordinates Points[], int pointsize)
{
	if (!PageOpen) NewPage();
	if (TextOpen) EndText();
	DynamicArray <coordinates> Coordinates;
	Coordinates.Length = pointsize;
	for (int i = 0; i < Coordinates.Length; i++) {
		Coordinates[i] = Points[i];
		}
	for (int i = 0; i < Coordinates.Length; i++) {
		Coordinates[i].X = 2.8346 * Coordinates[i].X;
		Coordinates[i].Y = 2.8346 * (PageHeight - Coordinates[i].Y);
		}
	PageBuffer->Add("% PolyBezier with " + AnsiString(pointsize) + " points");
	PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " m ");
    int max;
    for (int i = 1; i < Coordinates.Length - 2; i+=3) {
        PageBuffer->Add(DP(3,Coordinates[i].X) + " " + DP(3,Coordinates[i].Y) + " " + DP(3,Coordinates[i + 1].X)  + " " + DP(3,Coordinates[i + 1].Y) + " " + DP(3,Coordinates[i + 2].X)  + " " + DP(3,Coordinates[i + 2].Y) + " c ");
        max = i + 2;
        }
    PageBuffer->Add("S ");
    PenPosX = Coordinates[max].X;
    PenPosY = Coordinates[max].Y;

    Coordinates.Length = 0;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Polygon(coordinates Points[], int pointsize)
{
    if (!PageOpen) NewPage();
    if (TextOpen) EndText();
    DynamicArray <coordinates> Coordinates;
    Coordinates.Length = pointsize;
    for (int i = 0; i < Coordinates.Length; i++) {
        Coordinates[i] = Points[i];
        }
    for (int i = 0; i < Coordinates.Length; i++) {
        Coordinates[i].X = 2.8346 * Coordinates[i].X;
        Coordinates[i].Y = 2.8346 * (PageHeight - Coordinates[i].Y);
		}
	PageBuffer->Add("% Polygon with " + AnsiString(pointsize) + " points");
    PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " m ");
    for (int i = 1; i < Coordinates.Length; i++) {
        PageBuffer->Add(DP(3,Coordinates[i].X) + " " + DP(3,Coordinates[i].Y) + " l ");
        }
    PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " l ");
	//if (Brush->Style == bsSolid) {
	if (Brush.Style == bsSolid) {
		PageBuffer->Add("B ");
        } else {
        PageBuffer->Add("S ");
        }
    PenPosX = Coordinates[0].X;
    PenPosY = Coordinates[0].Y;

    Coordinates.Length = 0;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Polyline(coordinates Points[], int pointsize)
{
	PolyLine(Points,pointsize);
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::PolyLine(coordinates Points[], int pointsize)
{
    if (!PageOpen) NewPage();
    if (TextOpen) EndText();
    DynamicArray <coordinates> Coordinates;
    Coordinates.Length = pointsize;
    for (int i = 0; i < Coordinates.Length; i++) {
        Coordinates[i] = Points[i];
        }
    for (int i = 0; i < Coordinates.Length; i++) {
        Coordinates[i].X = 2.8346 * Coordinates[i].X;
        Coordinates[i].Y = 2.8346 * (PageHeight - Coordinates[i].Y);
        }
    PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " m ");
	int max;
	PageBuffer->Add("% PolyLine with " + AnsiString(pointsize) + " points");
    for (int i = 1; i < Coordinates.Length; i++) {
        PageBuffer->Add(DP(3,Coordinates[i].X) + " " + DP(3,Coordinates[i].Y) + " l ");
        max = i + 2;
        }
    PageBuffer->Add("S ");
    PenPosX = Coordinates[max].X;
    PenPosY = Coordinates[max].Y;

    Coordinates.Length = 0;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Rectangle( double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY)
{
    if (!PageOpen) NewPage();
    if (TextOpen) EndText();
    TColor oldcolour = PenColour;
    //adjust size so that outer edge of line is exactly on the defined boundary
    TopLeftX = TopLeftX + PenWidth/2.0;
    TopLeftY = TopLeftY + PenWidth/2.0;
    BottomRightX = BottomRightX - PenWidth/2.0;
    BottomRightY = BottomRightY - PenWidth/2.0;
	if (Pen->Style == psClear) PenColour = BrushColour;
	PageBuffer->Add("% Rectangle(" + AnsiString(TopLeftX) + "," + AnsiString(TopLeftY)  + "," + AnsiString(BottomRightX)  + "," + AnsiString(BottomRightY) + ")");
	//if (Brush->Style == bsSolid) {
	if (Brush.Style == bsSolid) {
		PageBuffer->Add(DP(3,2.8346 * TopLeftX) + " " + DP(3,2.8346 * (PageHeight - BottomRightY)) + " " + DP(3,2.8346 *(BottomRightX - TopLeftX)) + " " + DP(3,2.8346 *(BottomRightY - TopLeftY)) + " re ");
        PageBuffer->Add("B");
        } else {
        MoveTo(TopLeftX,TopLeftY);
        PageBuffer->Add(DP(3,2.8346 * TopLeftX) + " " + DP(3,2.8346 * (PageHeight - BottomRightY)) + " l ");
        PageBuffer->Add(DP(3,2.8346 * BottomRightX) + " " + DP(3,2.8346 * (PageHeight - BottomRightY)) + " l ");
        PageBuffer->Add(DP(3,2.8346 * BottomRightX) + " " + DP(3,2.8346 * (PageHeight - TopLeftY)) + " l ");
        PageBuffer->Add("h ");
        PageBuffer->Add("S ");
        }
    PenPosX = TopLeftX;
    PenPosY = TopLeftY;
    if (Pen->Style == psClear) PenColour = oldcolour;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::RoundRect(double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY, double EllipseWidth, double EllipseHeight)
{
    //Reproduces the TCanvas->RoundRect() method on a PDF page.
	if (!PageOpen) NewPage();
	if (TextOpen) EndText();
	TColor oldcolour = PenColour;
	if (Pen->Style == psClear) PenColour = BrushColour;
	//adjust size so that outer edge of line is exactly on the defined boundary
	TopLeftX = TopLeftX + PenWidth/2.0;
	TopLeftY = TopLeftY + PenWidth/2.0;
	BottomRightX = BottomRightX - PenWidth/2.0;
	BottomRightY = BottomRightY - PenWidth/2.0;
	EllipseWidth = EllipseWidth - PenWidth;
	EllipseHeight = EllipseHeight - PenWidth;

	DynamicArray <coordinates> Coordinates;
	Coordinates.Length = 16;
	coordinates delta, he;
	//the value below is half of the magic '1-K' for a bezier line where K = 0.5522847498 K=(4/3)*(sqrt(2)-1)
	delta.X = 0.2238576251 * EllipseWidth;
	delta.Y = 0.2238576251 * EllipseHeight;
	he.X = EllipseWidth/2.0;
	he.Y = EllipseHeight/2.0;
	//start bottom left corner arc
	Coordinates[0].X = TopLeftX;
	Coordinates[0].Y = BottomRightY - he.Y;
	Coordinates[1].X = TopLeftX;
	Coordinates[1].Y = BottomRightY - delta.Y;
	Coordinates[2].X = TopLeftX + delta.X;
	Coordinates[2].Y = BottomRightY;
	Coordinates[3].X = TopLeftX + he.X;
	Coordinates[3].Y = BottomRightY;
	//bottom right corner arc
	Coordinates[4].X = BottomRightX - he.X;
	Coordinates[4].Y = BottomRightY;
	Coordinates[5].X = BottomRightX - delta.X;
	Coordinates[5].Y = BottomRightY;
	Coordinates[6].X = BottomRightX;
	Coordinates[6].Y = BottomRightY - delta.Y;
	Coordinates[7].X = BottomRightX;
	Coordinates[7].Y = BottomRightY - he.Y;
	//top right corner arc
	Coordinates[8].X = BottomRightX;
	Coordinates[8].Y = TopLeftY + he.Y;
	Coordinates[9].X = BottomRightX;
	Coordinates[9].Y = TopLeftY + delta.Y;
	Coordinates[10].X = BottomRightX - delta.X;
	Coordinates[10].Y = TopLeftY;
	Coordinates[11].X = BottomRightX - he.X;
	Coordinates[11].Y = TopLeftY;
	//top left corner arc
	Coordinates[12].X = TopLeftX + he.X;
	Coordinates[12].Y = TopLeftY;
	Coordinates[13].X = TopLeftX + delta.X;
	Coordinates[13].Y = TopLeftY;
	Coordinates[14].X = TopLeftX;
	Coordinates[14].Y = TopLeftY + delta.Y;
	Coordinates[15].X = TopLeftX;
	Coordinates[15].Y = TopLeftY + he.Y;
	for (int i = 0; i < Coordinates.Length; i++) {
		Coordinates[i].X = 2.8346 * Coordinates[i].X;
        Coordinates[i].Y = 2.8346 * (PageHeight - Coordinates[i].Y);
		}
	PageBuffer->Add("% RoundRect(" + AnsiString(TopLeftX) + "," + AnsiString(TopLeftY)  + "," + AnsiString(BottomRightX)  + "," + AnsiString(BottomRightY)  + "," + AnsiString(EllipseWidth)  + "," + AnsiString(EllipseHeight) + ")");
	PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " m ");
	PageBuffer->Add(DP(3,Coordinates[1].X) + " " + DP(3,Coordinates[1].Y) + " " + DP(3,Coordinates[2].X)  + " " + DP(3,Coordinates[2].Y) + " " + DP(3,Coordinates[3].X)  + " " + DP(3,Coordinates[3].Y) + " c ");
	PageBuffer->Add(DP(3,Coordinates[4].X) + " " + DP(3,Coordinates[4].Y) + " l ");
	PageBuffer->Add(DP(3,Coordinates[5].X) + " " + DP(3,Coordinates[5].Y) + " " + DP(3,Coordinates[6].X)  + " " + DP(3,Coordinates[6].Y) + " " + DP(3,Coordinates[7].X)  + " " + DP(3,Coordinates[7].Y) + " c ");
	PageBuffer->Add(DP(3,Coordinates[8].X) + " " + DP(3,Coordinates[8].Y) + " l ");
	PageBuffer->Add(DP(3,Coordinates[9].X) + " " + DP(3,Coordinates[9].Y) + " " + DP(3,Coordinates[10].X)  + " " + DP(3,Coordinates[10].Y) + " " + DP(3,Coordinates[11].X)  + " " + DP(3,Coordinates[11].Y) + " c ");
	PageBuffer->Add(DP(3,Coordinates[12].X) + " " + DP(3,Coordinates[12].Y) + " l ");
	PageBuffer->Add(DP(3,Coordinates[13].X) + " " + DP(3,Coordinates[13].Y) + " " + DP(3,Coordinates[14].X)  + " " + DP(3,Coordinates[14].Y) + " " + DP(3,Coordinates[15].X)  + " " + DP(3,Coordinates[15].Y) + " c ");
	PageBuffer->Add(DP(3,Coordinates[0].X) + " " + DP(3,Coordinates[0].Y) + " l ");
	//if (Brush->Style == bsSolid) {
	if (Brush.Style == bsSolid) {
		PageBuffer->Add("B ");
        } else {
		PageBuffer->Add("S ");
        }
	PenPosX = TopLeftX;
    PenPosY = (BottomRightY + TopLeftY)/2.0;
	if (Pen->Style == psClear) PenColour = oldcolour;

	Coordinates.Length = 0;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPDFObject::ClippingArea(double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY)
{
    AnsiString text;

    text = DP(2,2.8346 * (BottomRightX - TopLeftX)) + " ";        //X Scale
    text = text + "0 0 ";
    text = text + DP(2,2.8346 * (BottomRightY - TopLeftY)) + " "; //Y Scale
    text = text + DP(2,2.8346 * TopLeftX) + " ";
    text = text + DP(2,2.8346 * (PageHeight - BottomRightY)) + " ";
    text = text + "cm ";


    return text;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPDFObject::GetPenColour(void)
{
    rgbcomponents Colour = ExtractRGBcomponents(Pen->Color);
    AnsiString text = DP(2,Colour.R) + " " + DP(2,Colour.G) + " " + DP(2,Colour.B) + " RG ";

    return text;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPDFObject::GetBrushColour(void)
{
	//rgbcomponents Colour = ExtractRGBcomponents(Brush->Color);
	rgbcomponents Colour = ExtractRGBcomponents(Brush.Color);
	AnsiString text = DP(2,Colour.R) + " " + DP(2,Colour.G) + " " + DP(2,Colour.B) + " rg ";

    return text;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPDFObject::GetTextColour(void)
{
	//rgbcomponents Colour = ExtractRGBcomponents(Font->Color);
	rgbcomponents Colour = ExtractRGBcomponents(Font.Color);
	AnsiString text = DP(2,Colour.R) + " " + DP(2,Colour.G) + " " + DP(2,Colour.B) + " rg ";

    return text;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPDFObject::SanitiseText(AnsiString text)
{
	AnsiString Result;

	//check for unbalanced brackets & character values exceeding 127
	int openbrackets = 0;
	int closebrackets = 0;
	AnsiString temp;
	char a;
	for (int i = 1; i < text.Length() + 1; i++) {
		temp = text.SubString(i,1);
		if (temp == "(") openbrackets++;
		if (temp == ")") closebrackets++;
		a = 0;
		while ( a <= 255 && AnsiString(a) != temp) a++;
		if ( (int)(a) < 0) {
			Result = Result + "\\" + Octal(a);
			} else {
			Result = Result + temp;
            if (temp == "\\") Result += temp;
			}
		}

	if (openbrackets != closebrackets) {
		if (openbrackets > closebrackets) {
			int numberofreplacements = openbrackets - closebrackets;
			int pointer = Result.Length() + 1;
			while (numberofreplacements > 0 && pointer > 0) {
				if (Result.SubString(pointer,1) == "(") {
					Result = LeftString(Result,pointer - 1) + "\\" + RightString(Result,Result.Length()-pointer + 1);
					numberofreplacements--;
					}
				pointer--;
				}
			} else {
			int numberofreplacements = closebrackets - openbrackets;
			int pointer = Result.Length();
			while (numberofreplacements > 0 && pointer > 0) {
				if (Result.SubString(pointer,1) == ")") {
					Result = LeftString(Result,pointer-1) + "\\" + RightString(Result,Result.Length()-pointer);
					numberofreplacements--;
					}
				pointer--;
				}
			}
		}

	return Result;
}
//---------------------------------------------------------------------------

AnsiString __fastcall TPDFObject::Octal(char a)
{
	//converts char a into an octal string
	int c1 = (int)(a);
    if (c1 < 0) c1 += 256;
	int maxb = c1 >> 6;
	c1 = c1 - (maxb << 6);
	int midb = c1 >> 3;
	c1 = c1 - (midb << 3);

	AnsiString Result;
	if (maxb) Result = AnsiString(maxb);
	if (maxb || midb) Result += AnsiString(midb);
	Result += AnsiString(c1);

	return Result;
}
//---------------------------------------------------------------------------

int __fastcall TPDFObject::AddResourceImage(Graphics::TBitmap *Bitmap)
{
    //This [very] public method adds a graphic to the resource list.
    //It is added to the ImageResourceList for later use in the program
    int returnvalue = nextobjectnumber;
    AddObject(nextobjectnumber,0,false);
    //set up resource reference
    int pointer = ImageResourceList.Length;
    ImageResourceList.Length++;
    ImageResourceList[pointer].Name = "Image" + AnsiString(ImageResourceList.Length);
    ImageResourceList[pointer].objectnumber = nextobjectnumber;
    ImageResourceList[pointer].generation = 0;
    //write the resource image
    WriteLine(AnsiString(nextobjectnumber) + " 0 obj ");
    WriteLine("<< ");
    WriteLine("/Type /XObject ");
    WriteLine("/Subtype /Image ");
    WriteLine("/Width " + AnsiString(Bitmap->Width) + " ");
    WriteLine("/Height " + AnsiString(Bitmap->Height) + " ");
    WriteLine("/ColorSpace /DeviceRGB ");
    WriteLine("/BitsPerComponent 8 ");
    //WriteLine("/Length " + AnsiString(nextobjectnumber + 1) + " 0 R ");
    WriteLine("/Length " + AnsiString(3 * Bitmap->Width * Bitmap->Height) + " ");
    WriteLine(">> ");
    WriteLine("stream");
    //int start = streampointer;
    rgb pixelcolour;
    char c[3];
    for (int j = 0; j < Bitmap->Height; j++) {
        for (int i = 0; i < Bitmap->Width; i++) {
            pixelcolour = ExtractRGB(Bitmap->Canvas->Pixels[i][j]);
            c[0] = ((char)(pixelcolour.R));
            c[1] = ((char)(pixelcolour.G));
            c[2] = ((char)(pixelcolour.B));
            fwrite(c,3,1,output);
            streampointer += 3;
            }
        }
    //int end = streampointer;
    AnsiString text = "endstream";
    fwrite(text.c_str(),strlen(text.c_str()),1,output);
    streampointer += strlen(text.c_str());
    CRLF();
    WriteLine("endobj");
    nextobjectnumber++;

    return returnvalue;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::UseResourceImage(int ObjectNumber, double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY)
{
    //This routine causes an existing resource to be displayed
    //Will only work if the resource has been added to the list.

    //first find if the resource has been added
    bool found = false;
    int pointer = 0;

    while (!found && pointer < ImageResourceList.Length) {
        if (ImageResourceList[pointer].objectnumber == ObjectNumber) {
            found = true;
            pointer--;
            }
        pointer++;
        }

    if (found) {
		//check if a page has been started. If not - start page.
		if (!PageOpen) NewPage();
		//if text open, the close it.
		if (TextOpen) EndText();
		//OK, the resource image exists and can be used.
        //now check if already referred to
        int pointer2 = 0;
        found = false;
        int arraypointer = nextpagenumber - 2;
        while (!found && pointer2 < PageList[arraypointer].ImageResource.Length) {
            if (PageList[arraypointer].ImageResource[pointer2].objectnumber == ObjectNumber) {
                found = true;
                pointer2--;
                }
            pointer2++;
            }
        if (!found) {
            //not found, so make page list refer to image
            PageList[arraypointer].ImageResource.Length++;
            PageList[arraypointer].ImageResource[PageList[arraypointer].ImageResource.Length - 1].objectnumber = ObjectNumber;
            PageList[arraypointer].ImageResource[PageList[arraypointer].ImageResource.Length - 1].generation = ImageResourceList[pointer].generation;
            }
        PageList[arraypointer].resourceImageC = true;
        //now add the referenced image to the page contents
        double userspacewidth = (BottomRightX - TopLeftX)* 2.8346;
		double userspaceheight = (BottomRightY - TopLeftY)* 2.8346;
		PageBuffer->Add("% UseResourceImage");
        PageBuffer->Add("q ");
        PageBuffer->Add(DP(3, userspacewidth) + " 0 0 " + DP(3,userspaceheight) + " " + DP(3,TopLeftX * 2.8346) + " " + DP(3,2.8346 * (PageHeight-BottomRightY)) + " cm ");
        PageBuffer->Add("/"+ImageResourceList[pointer].Name + " Do ");
        PageBuffer->Add("Q");
        }
}
//---------------------------------------------------------------------------

int __fastcall TPDFObject::Draw(Graphics::TBitmap *Bitmap, double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY)
{
    //Draw uses the standard AddResource and UseResource commands to draw an image.
    //returns the object number of the graphic so it can be re-used with
    //another UseResourceImage() if need be.
    if (!PageOpen) NewPage();
    if (TextOpen) EndText();
    int objectnumber = AddResourceImage(Bitmap);
    UseResourceImage(objectnumber,TopLeftX,TopLeftY,BottomRightX,BottomRightY);

    return objectnumber;
}
//---------------------------------------------------------------------------

double __fastcall TPDFObject::TextHeight(void)
{
	//the text height is returned as the point height (72 points per inch) converted into mm at 2.8346 points per mm.
	//This is the 'Official' height of the font, and does not take into account any descenders below the font line.
	//return (((double)(Font->Size)) / 2.8346);
	return (((double)(Font.Size)) / 2.8346);
}
//---------------------------------------------------------------------------

double __fastcall TPDFObject::TextHeight(AnsiString text)
{
    double returnvalue = 0.0;

    Graphics::TBitmap *Bitmap = new Graphics::TBitmap();
	Bitmap->Canvas->Font->Name = Font.Name;
	Bitmap->Canvas->Font->Size = Font.Size;
	if (FontStyle.Contains(fsBold)) Bitmap->Canvas->Font->Style << fsBold;
	if (FontStyle.Contains(fsItalic)) Bitmap->Canvas->Font->Style << fsItalic;
	if (FontStyle.Contains(fsUnderline)) Bitmap->Canvas->Font->Style << fsUnderline;
	if (FontStyle.Contains(fsStrikeOut)) Bitmap->Canvas->Font->Style << fsStrikeOut;
	Bitmap->Canvas->Font->Size = 10*Bitmap->Canvas->Font->Size;
	double width =  (double)(Bitmap->Canvas->TextHeight(text));
    Bitmap->Canvas->Font->Name = "Arial";
    Bitmap->Canvas->Font->Size = 100;
	//'Hi There' in 100 point Arial is 135.170mm wide. This is from tests with a HP Laserjet 5100 printer at 600 dpi horizontally and the reported width in pixels, converted into mm.
    double referencewidth =  (double)(Bitmap->Canvas->TextWidth("Hi There"));
    if (referencewidth > 0.0) returnvalue = 13.517 * width/referencewidth;

    delete Bitmap;

    return returnvalue;
}
//---------------------------------------------------------------------------

double __fastcall TPDFObject::TextWidth(AnsiString text)
{
    double returnvalue = 0.0;

    Graphics::TBitmap *Bitmap = new Graphics::TBitmap();
	//Bitmap->Canvas->Font->Assign(Font);
	if (Font.Name.Pos("Times")) {
		Bitmap->Canvas->Font->Name = "Times New Roman";
		} else {
		if (Font.Name.Pos("Helvetica")) {
			Bitmap->Canvas->Font->Name = "Arial";
			} else {
			if (Font.Name.Pos("Courier")) {
				Bitmap->Canvas->Font->Name = "Courier New";
				} else {
                Bitmap->Canvas->Font->Name = Font.Name;
                }
            }
		}

	Bitmap->Canvas->Font->Size = Font.Size;
	if (FontStyle.Contains(fsBold)) Bitmap->Canvas->Font->Style << fsBold;
	if (FontStyle.Contains(fsItalic)) Bitmap->Canvas->Font->Style << fsItalic;
	if (FontStyle.Contains(fsUnderline)) Bitmap->Canvas->Font->Style << fsUnderline;
	if (FontStyle.Contains(fsStrikeOut)) Bitmap->Canvas->Font->Style << fsStrikeOut;
	Bitmap->Canvas->Font->Size = 10*Bitmap->Canvas->Font->Size;
    double width =  (double)(Bitmap->Canvas->TextWidth(text));
    Bitmap->Canvas->Font->Name = "Arial";
    Bitmap->Canvas->Font->Size = 100;
	//'Hi There' in 100 point Arial is 135.170mm wide. This is from tests with a HP Laserjet 5100 printer at 600 dpi horizontally and the reported width in pixels, converted into mm.
    double referencewidth =  (double)(Bitmap->Canvas->TextWidth("Hi There"));
    if (referencewidth > 0.0) returnvalue = 13.517 * width/referencewidth;

    delete Bitmap;

    return returnvalue;
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::TextOut(AnsiString text, double TopLeftX, double TopLeftY)
{
	//oops! Obsolete function.
	TextOut(TopLeftX,TopLeftY,text);
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::TextOut(double TopLeftX, double TopLeftY, AnsiString text)
{
	if (!PageOpen) NewPage();
	if (!TextOpen) BeginText();
    if (PageOpen && TextOpen) {
        //check 1, has the current font been added to the resource list?
        //if no, add it.
        bool found = false;
        bool Type1Font = false;
        int pointer = 0;
		//if (Font->Name == "Courier" || Font->Name == "Helvetica") {
		if (Font.Name == "Courier" || Font.Name == "Helvetica") {
			if (FontStyle.Contains(fsBold) || FontStyle.Contains(fsItalic)) {
				//Font->Name = Font->Name + "-";
				Font.Name = Font.Name + "-";
				}
            if (FontStyle.Contains(fsBold)) {
				//Font->Name = Font->Name + "Bold";
				Font.Name = Font.Name + "Bold";
				}
            if (FontStyle.Contains(fsItalic)) {
				//Font->Name = Font->Name + "Oblique";
				Font.Name = Font.Name + "Oblique";
				}
            Type1Font = true;
            }
		//if (Font->Name == "Times") {
		if (Font.Name == "Times") {
			if (FontStyle.Contains(fsBold) || FontStyle.Contains(fsItalic)) {
				//Font->Name = Font->Name + "-";
				Font.Name = Font.Name + "-";
				} else {
				//Font->Name = Font->Name + "-Roman";
				Font.Name = Font.Name + "-Roman";
				}
            if (FontStyle.Contains(fsBold)) {
				//Font->Name = Font->Name + "Bold";
				Font.Name = Font.Name + "Bold";
				}
            if (FontStyle.Contains(fsItalic)) {
				//Font->Name = Font->Name + "Italic";
				Font.Name = Font.Name + "Italic";
				}
            Type1Font = true;
            }
		//if (Font->Name == "Symbol" || Font->Name == "ZapfDingbats") {
		if (Font.Name == "Symbol" || Font.Name == "ZapfDingbats") {
			Type1Font = true;
            }
        while (!found && pointer < FontResourceList.Length) {
			//if (FontResourceList[pointer].FontName == Font->Name
			if (FontResourceList[pointer].FontName == Font.Name
				&& FontResourceList[pointer].bold == FontStyle.Contains(fsBold)
                && FontResourceList[pointer].italic == FontStyle.Contains(fsItalic)
                && FontResourceList[pointer].underlined == FontStyle.Contains(fsUnderline)
                && FontResourceList[pointer].strikeout == FontStyle.Contains(fsStrikeOut)
                ) {
                found = true;
                pointer--;
                }
            pointer++;
            }
        if (!found) WriteFontResource(Type1Font);
        //pointer will now be correctly pointing to the last entry if the font had to be added to the resource list

        //check 2, has the current font been added to the current page resource list?
        //if no, add it.
        int objectnumber = FontResourceList[pointer].objectnumber;
        found = false;
        int pointer2 = 0;
        //int pagelistpointer = PageList.Length - 1;
        int pagelistpointer = nextpagenumber - 2;
        while (!found && pointer2 < PageList[pagelistpointer].FontResource.Length) {
            if (PageList[pagelistpointer].FontResource[pointer2].objectnumber == objectnumber) {
                found = true;
                pointer2--;
                }
            pointer2++;
            }
        if (!found) {
            PageList[pagelistpointer].FontResource.Length++;
            PageList[pagelistpointer].FontResource[pointer2].objectnumber = objectnumber;
            PageList[pagelistpointer].FontResource[pointer2].generation = FontResourceList[pointer].generation;
            PageList[pagelistpointer].resourceText = true;
            }

        //sanitise text so it won't give spurious results
		text = SanitiseText(text);


        //Now, at last, we can get around to actually writing the text.
		//specify font name & point size
		PageBuffer->Add("/" + FontResourceList[pointer].Name + " " + AnsiString(Font.Size) + " Tf ");
		//specify position
		PageBuffer->Add("1 0 0 1 " + DP(3,TopLeftX * 2.8346) + " " + DP(3,(PageHeight - TopLeftY) * 2.8346 - ((double)(Font.Size))) + " Tm ");
		//specify colour
        PageBuffer->Add(GetTextColour());
        //write the text
        PageBuffer->Add("(" + text + ") Tj ");
		if (FontStyle.Contains(fsUnderline) || FontStyle.Contains(fsStrikeOut)) {
			//underlining and strikeout is done by drawing the lines rather than in the font.
            //also do not want the lining to disturb the line drawing colour or the pen position, so restore them at the end
			double textwidth = TextWidth(text);
			EndText();
			bool setcolour = false;
			TColor oldcolour = PenColour;
			double oldpenposx = fPenPosX;
			double oldpenposy = fPenPosY;
			if (PenColour != Font.Color) {
				setcolour = true;
				PenColour = Font.Color;
				}
			PageBuffer->Add(DP(2,Font.Size/15.0) + " w ");
			if (FontStyle.Contains(fsUnderline)) {
				MoveTo(TopLeftX,TopLeftY + ((double)(Font.Size) * 1.1 /*+ 2.0*/)/2.8346);
				LineTo(TopLeftX + textwidth,TopLeftY + ((double)(Font.Size) * 1.1 /*+ 2.0*/)/2.8346);
				}
			if (FontStyle.Contains(fsStrikeOut)) {
				MoveTo(TopLeftX,TopLeftY + 3.0 * Font.Size/(4.0*2.8346));
				LineTo(TopLeftX + textwidth,TopLeftY + 3.0 * Font.Size/(4.0*2.8346));
				}
			PageBuffer->Add(DP(2,PenWidth*2.8346) + " w ");
			if (setcolour) {
				PenColour = oldcolour;
				}
			//change fPenPos_, as don't want to trigger any actions associated
			//with setting the properties should they be added in the future.
			fPenPosX = oldpenposx;
			fPenPosY = oldpenposy;
			}
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::BeginText(void)
{
    if (PageOpen) {
        TextOpen = true;
        //begin text
        PageBuffer->Add("BT ");
        }
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::EndText(void)
{
    if (PageOpen && TextOpen) {
        TextOpen = false;
        //begin text
        PageBuffer->Add("ET ");
        }
}
//---------------------------------------------------------------------------


void __fastcall TPDFObject::WriteFontResource(bool Type1Font)
{
    if (Type1Font) {
        //writes the font object. It's a standard built in font
        int pointer = FontResourceList.Length;
        FontResourceList.Length++;
		FontResourceList[pointer].Name = "Font" + AnsiString(FontResourceList.Length);
		//FontResourceList[pointer].FontName = Font->Name;
		FontResourceList[pointer].FontName = Font.Name;
		FontResourceList[pointer].bold = FontStyle.Contains(fsBold);
        FontResourceList[pointer].italic = FontStyle.Contains(fsItalic);
        FontResourceList[pointer].underlined = false;
        FontResourceList[pointer].strikeout = false;
        FontResourceList[pointer].objectnumber = nextobjectnumber;
        FontResourceList[pointer].generation = 0;
        AddObject(nextobjectnumber,0,false);
        WriteLine(AnsiString(FontResourceList[pointer].objectnumber) + " " + AnsiString(FontResourceList[pointer].generation) + " obj ");
        WriteLine("<< ");
        WriteLine("/Type /Font ");
        WriteLine("/Subtype /Type1");
        WriteLine("/Name /" + FontResourceList[pointer].Name + " ");
		//WriteLine("/BaseFont /" + Font->Name + " ");
		WriteLine("/BaseFont /" + Font.Name + " ");
		WriteLine("/Encoding /WinAnsiEncoding");

        WriteLine(">> ");
        WriteLine("endobj ");
        nextobjectnumber++;
        } else {
		//It's a Truetype Font, and a lot more work needs to be done to embed all the font information, and I'm pretty certain I'm not up to this.
		//bother.
		//bother, Bother, BOTHER
		//Gaah!
		//For the moment I have copped out, and treated it as a non-embedded font on the assumption that it'll be found on the host machine anyway.

		//writes the font object.
		int pointer = FontResourceList.Length;
		FontResourceList.Length++;
        FontResourceList[pointer].Name = "Font" + AnsiString(FontResourceList.Length);
		FontResourceList[pointer].FontName = Font.Name;
		FontResourceList[pointer].bold = FontStyle.Contains(fsBold);
        FontResourceList[pointer].italic = FontStyle.Contains(fsItalic);
        FontResourceList[pointer].underlined = FontStyle.Contains(fsUnderline);
        FontResourceList[pointer].strikeout = FontStyle.Contains(fsStrikeOut);
        FontResourceList[pointer].objectnumber = nextobjectnumber;
        FontResourceList[pointer].generation = 0;
        AddObject(nextobjectnumber,0,false);
        WriteLine(AnsiString(FontResourceList[pointer].objectnumber) + " " + AnsiString(FontResourceList[pointer].generation) + " obj ");
        WriteLine("<< ");
        WriteLine("/Type /Font ");
        WriteLine("/Subtype /TrueType");
        WriteLine("/Name /" + FontResourceList[pointer].Name + " ");
		WriteLine("/BaseFont /" + Font.Name + " ");
		WriteLine("/Encoding /WinAnsiEncoding");
		WriteLine(">> ");
        WriteLine("endobj ");
		nextobjectnumber++;
        }

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::URL(AnsiString DisplayText, AnsiString Destination, double TopLeftX, double TopLeftY)
{
	//show URL even if hyperlink fails
    TextOut(DisplayText, TopLeftX, TopLeftY);
    double BottomRightX = TopLeftX + TextWidth(DisplayText);
    double BottomRightY = TopLeftY + TextHeight();

    //Now to create the annotation object.
    //these objects are only called once, so no need to make them referenceable
    //int pointer = PageList.Length - 1;
    int pointer = nextpagenumber - 2;
    int pointer2 = PageList[pointer].AnnotsResource.Length;
    PageList[pointer].AnnotsResource.Length++;
    PageList[pointer].AnnotsResource[pointer2].objectnumber = nextobjectnumber;
    PageList[pointer].AnnotsResource[pointer2].generation = 0;

    AddObject(nextobjectnumber,0,false);
    WriteLine(AnsiString(PageList[pointer].AnnotsResource[pointer2].objectnumber) + " " + AnsiString(PageList[pointer].AnnotsResource[pointer2].generation) + " obj ");
    WriteLine("<< ");
    WriteLine("/Subtype /Link ");
    WriteLine("/Rect [ " + DP(2,TopLeftX * 2.8346) + " " + DP(2,2.8346 * (PageHeight-BottomRightY)) + " " + DP(2,BottomRightX * 2.8346) + " " + DP(2,2.8346*(PageHeight - TopLeftY)) + " ] ");
    WriteLine("/BS <</W 0 >>");
    WriteLine("/F 4 ");
    WriteLine("/A <</Type /Action /S /URI /URI (" + Destination + ") >>");
	WriteLine(">> ");
    WriteLine("endobj ");

    nextobjectnumber++;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::PageLink(AnsiString DisplayText, int DestinationPage, double TopLeftX, double TopLeftY)
{
    int arraypointer = 0;
    bool found = false;
    while (!found && arraypointer < PageList.Length) {
        if (PageList[arraypointer].pagenumber == DestinationPage) {
            found = true;
            arraypointer--;
            }
        arraypointer++;
        }

    //show the text anyway even if the link cannot be created.
    TextOut(DisplayText, TopLeftX, TopLeftY);

    if (!found) {
        MakeEnoughPages(DestinationPage);
        PageList[DestinationPage-1].objectnumber = nextobjectnumber;
        nextobjectnumber++;
        arraypointer = DestinationPage-1;
        }

    //link can now be created
    double BottomRightX = TopLeftX + TextWidth(DisplayText);
    double BottomRightY = TopLeftY + TextHeight();

    //Now to create the annotation object.
    //these objects are only called once, so no need to make them referenceable
    int pointer = nextpagenumber - 2;
    int pointer2 = PageList[pointer].AnnotsResource.Length;
    PageList[pointer].AnnotsResource.Length++;
    PageList[pointer].AnnotsResource[pointer2].objectnumber = nextobjectnumber;
    PageList[pointer].AnnotsResource[pointer2].generation = 0;

    AddObject(nextobjectnumber,0,false);
    WriteLine(AnsiString(PageList[pointer].AnnotsResource[pointer2].objectnumber) + " " + AnsiString(PageList[pointer].AnnotsResource[pointer2].generation) + " obj ");
    WriteLine("<< ");
    WriteLine("/Subtype /Link ");
    WriteLine("/Rect [ " + DP(2,TopLeftX * 2.8346) + " " + DP(2,2.8346 * (PageHeight-BottomRightY)) + " " + DP(2,BottomRightX * 2.8346) + " " + DP(2,2.8346*(PageHeight - TopLeftY)) + " ] ");
    WriteLine("/BS <</W 0 >>");
    WriteLine("/F 4 ");
    WriteLine("/A <</Type /Action /S /GoTo /D [ " + AnsiString(PageList[arraypointer].objectnumber) + " " + AnsiString(PageList[arraypointer].generation) + " R /XYZ 0 " + DP(2,PageHeight * 2.8346) + " 0 ] >>");
	WriteLine(">> ");
    WriteLine("endobj ");

    nextobjectnumber++;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::HotSpot(int DestinationPage, double TopLeftX, double TopLeftY, double BottomRightX, double BottomRightY)
{
    int arraypointer = 0;
    bool found = false;
    while (!found && arraypointer < PageList.Length) {
        if (PageList[arraypointer].pagenumber == DestinationPage) {
            found = true;
            arraypointer--;
            }
        arraypointer++;
        }

    if (!found) {
        MakeEnoughPages(DestinationPage);
        PageList[DestinationPage-1].objectnumber = nextobjectnumber;
        nextobjectnumber++;
        arraypointer = DestinationPage-1;
        }

    //link can now be created
    //Now to create the annotation object.
    //these objects are only called once, so no need to make them referenceable
    int pointer = nextpagenumber - 2;
    int pointer2 = PageList[pointer].AnnotsResource.Length;
    PageList[pointer].AnnotsResource.Length++;
    PageList[pointer].AnnotsResource[pointer2].objectnumber = nextobjectnumber;
    PageList[pointer].AnnotsResource[pointer2].generation = 0;

    AddObject(nextobjectnumber,0,false);
    WriteLine(AnsiString(PageList[pointer].AnnotsResource[pointer2].objectnumber) + " " + AnsiString(PageList[pointer].AnnotsResource[pointer2].generation) + " obj ");
    WriteLine("<< ");
    WriteLine("/Subtype /Link ");
    WriteLine("/Rect [ " + DP(2,TopLeftX * 2.8346) + " " + DP(2,2.8346 * (PageHeight-BottomRightY)) + " " + DP(2,BottomRightX * 2.8346) + " " + DP(2,2.8346*(PageHeight - TopLeftY)) + " ] ");
    WriteLine("/BS <</W 0 >>");
    WriteLine("/F 4 ");
    WriteLine("/A <</Type /Action /S /GoTo /D [ " + AnsiString(PageList[arraypointer].objectnumber) + " " + AnsiString(PageList[arraypointer].generation) + " R /XYZ 0 " + DP(2,PageHeight * 2.8346) + " 0 ] >>");
    WriteLine(">> ");
    WriteLine("endobj ");

    nextobjectnumber++;

}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::Comment(AnsiString CommentText)
{
	//adds the comment to the page buffer
	if (!PageOpen) NewPage();
	PageBuffer->Add("% " + CommentText);
}
//---------------------------------------------------------------------------

void __fastcall TPDFObject::MakeEnoughPages(int PageNumber)
{
    if (PageNumber > PageList.Length) {
        int pointer = PageList.Length;
        PageList.Length = PageNumber;
        for (int i = pointer; i < PageList.Length; i++){
            PageList[i].pagenumber = i-1;
            PageList[i].objectnumber = 0;
            PageList[i].generation = 0;
            PageList[i].contentsobjectnumber = 0;
            PageList[i].contentsgeneration = 0;
            PageList[i].used = false;
            PageList[i].resourceImageB = false;
            PageList[i].resourceImageC = false;
            PageList[i].resourceText = false;
			PageList[i].FontResource.Length = 0;
            PageList[i].ImageResource.Length = 0;
            PageList[i].AnnotsResource.Length = 0;
            }
        }
}
//---------------------------------------------------------------------------






